#    (c) Copyright 2026 Hewlett Packard Enterprise Development LP
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import unittest
from unittest.mock import Mock, MagicMock, patch
import sys
import os

# Add the src directory to the path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from hpe_storage_flowkit_py.v3.src.workflows.cpg import CpgWorkflow
from hpe_storage_flowkit_py.v3.src.core.exceptions import HPEStorageException
from hpe_storage_flowkit_py.v3.src.validators.cpg_validator import validate_cpg_params


class MockSessionManager:
    """Mock session manager for testing"""
    def __init__(self):
        self.rest_client = MockRestClient()


class MockTaskManager:
    """Mock task manager for testing"""
    def __init__(self, session_mgr=None):
        self.session_mgr = session_mgr
    
    def waitForTaskToEnd(self, task_id, poll_rate_secs=15):
        """Mock wait for task completion"""
        return {'status': 'completed', 'taskId': task_id}


class MockRestClient:
    """Mock REST client for testing"""
    def __init__(self):
        self.responses = {}
        self.posted_data = []
        self.deleted_endpoints = []
        self.existing_cpgs = set()  # Track which CPGs "exist"
        
    def post(self, endpoint, payload):
        self.posted_data.append({'endpoint': endpoint, 'payload': payload})
        name = payload.get('name')
        if name:
            self.existing_cpgs.add(name)
        return {'status': 'created', 'name': name, 'payload': payload}
    
    def delete(self, endpoint):
        self.deleted_endpoints.append(endpoint)
        return {'status': 'deleted', 'endpoint': endpoint}
    
    def get(self, endpoint, headers=None):
        if '/cpgs?name=' in endpoint:
            name = endpoint.split('=')[1]
            # Return empty list if CPG doesn't exist (for existence check)
            if name in self.existing_cpgs:
                return [{'uid': f'cpg-{name}-uid', 'name': name}]
            else:
                return []
        elif endpoint == '/cpgs':
            return {'members': [
                {'uid': 'cpg-1', 'name': 'CPG1'},
                {'uid': 'cpg-2', 'name': 'CPG2'}
            ]}
        return {}


class TestCpgWorkflow(unittest.TestCase):
    """Comprehensive unit tests for CPG workflow"""
    
    def setUp(self):
        """Set up test fixtures"""
        self.session_mgr = MockSessionManager()
        self.task_mgr = MockTaskManager(self.session_mgr)
        self.workflow = CpgWorkflow(self.session_mgr, self.task_mgr)
    
    def _simulate_cpg_exists(self, name):
        """Helper to mark a CPG as existing"""
        self.session_mgr.rest_client.existing_cpgs.add(name)
    
    def test_create_cpg_basic(self):
        """Test creating CPG with only name"""
        result = self.workflow.create_cpg('test_cpg')
        self.assertEqual(result['status'], 'created')
        self.assertEqual(result['name'], 'test_cpg')
    
    def test_create_cpg_with_domain(self):
        """Test creating CPG with domain"""
        result = self.workflow.create_cpg('test_cpg', domain='test_domain')
        
        posted = self.session_mgr.rest_client.posted_data[-1]
        self.assertIn('domain', posted['payload'])
        self.assertEqual(posted['payload']['domain'], 'test_domain')
    
    def test_create_cpg_with_growth_size(self):
        """Test creating CPG with growthSizeMiB"""
        result = self.workflow.create_cpg('test_cpg', growthSizeMiB=10240)
        
        posted = self.session_mgr.rest_client.posted_data[-1]
        self.assertIn('growthSizeMiB', posted['payload'])
        self.assertEqual(posted['payload']['growthSizeMiB'], 10240)
    
    def test_create_cpg_with_growth_limit(self):
        """Test creating CPG with growthLimitMiB"""
        result = self.workflow.create_cpg('test_cpg', growthLimitMiB=102400)
        
        posted = self.session_mgr.rest_client.posted_data[-1]
        self.assertIn('growthLimitMiB', posted['payload'])
        self.assertEqual(posted['payload']['growthLimitMiB'], 102400)
    
    def test_create_cpg_with_growth_warning(self):
        """Test creating CPG with growthWarnMiB"""
        result = self.workflow.create_cpg('test_cpg', growthWarnMiB=51200)
        
        posted = self.session_mgr.rest_client.posted_data[-1]
        self.assertIn('growthWarnMiB', posted['payload'])
        self.assertEqual(posted['payload']['growthWarnMiB'], 51200)
    
    def test_create_cpg_with_ha(self):
        """Test creating CPG with high availability setting"""
        result = self.workflow.create_cpg('test_cpg', ha='HAJBOD_JBOD')
        
        posted = self.session_mgr.rest_client.posted_data[-1]
        self.assertIn('ha', posted['payload'])
        self.assertEqual(posted['payload']['ha'], 'HAJBOD_JBOD')
    
    def test_create_cpg_with_cage(self):
        """Test creating CPG with cage parameter"""
        result = self.workflow.create_cpg('test_cpg', cage='cage1')
        
        posted = self.session_mgr.rest_client.posted_data[-1]
        self.assertIn('cage', posted['payload'])
        self.assertEqual(posted['payload']['cage'], 'cage1')
    
    def test_create_cpg_with_position(self):
        """Test creating CPG with position parameter"""
        result = self.workflow.create_cpg('test_cpg', position='POSITION_FIRST')
        
        posted = self.session_mgr.rest_client.posted_data[-1]
        self.assertIn('position', posted['payload'])
        self.assertEqual(posted['payload']['position'], 'POSITION_FIRST')
    
    def test_create_cpg_with_key_value_pairs(self):
        """Test creating CPG with keyValuePairs"""
        kvp = {'env': 'production', 'owner': 'team1'}
        result = self.workflow.create_cpg('test_cpg', keyValuePairs=kvp)
        
        posted = self.session_mgr.rest_client.posted_data[-1]
        self.assertIn('keyValuePairs', posted['payload'])
        self.assertEqual(posted['payload']['keyValuePairs'], kvp)
    
    def test_create_cpg_with_all_parameters(self):
        """Test creating CPG with all possible parameters"""
        result = self.workflow.create_cpg(
            'test_cpg',
            domain='production',
            growthSizeMiB=10240,
            growthLimitMiB=102400,
            growthWarnMiB=81920,
            ha='HAJBOD_DISK',
            cage='cage1',
            position='POSITION_FIRST',
            keyValuePairs={'env': 'prod'}
        )
        
        posted = self.session_mgr.rest_client.posted_data[-1]
        payload = posted['payload']
        
        self.assertEqual(payload['name'], 'test_cpg')
        self.assertEqual(payload['domain'], 'production')
        self.assertEqual(payload['growthSizeMiB'], 10240)
        self.assertEqual(payload['growthLimitMiB'], 102400)
        self.assertEqual(payload['growthWarnMiB'], 81920)
        self.assertEqual(payload['ha'], 'HAJBOD_DISK')
        self.assertEqual(payload['cage'], 'cage1')
        self.assertEqual(payload['position'], 'POSITION_FIRST')
        self.assertEqual(payload['keyValuePairs'], {'env': 'prod'})
    
    def test_list_cpgs(self):
        """Test listing all CPGs"""
        result = self.workflow.list_cpgs()
        
        self.assertIn('members', result)
        self.assertEqual(len(result['members']), 2)
        self.assertEqual(result['members'][0]['name'], 'CPG1')
        self.assertEqual(result['members'][1]['name'], 'CPG2')
    
    def test_create_cpg_invalid_name(self):
        """Test creating CPG with invalid name should raise error"""
        with self.assertRaises(Exception):
            self.workflow.create_cpg('')
    
    def test_create_cpg_none_name(self):
        """Test creating CPG with None name should raise error"""
        with self.assertRaises(Exception):
            self.workflow.create_cpg(None)
    
    def test_delete_cpg_invalid_name(self):
        """Test deleting CPG with invalid name should raise error"""
        with self.assertRaises(Exception):
            self.workflow.delete_cpg('')
    
    def test_create_cpg_with_unknown_parameter(self):
        """Test creating CPG with unknown parameter should raise error"""
        with self.assertRaises(ValueError):
            self.workflow.create_cpg('test_cpg', unknown_param='value')
    
    def test_create_cpg_multiple_kwargs(self):
        """Test creating CPG with multiple kwargs"""
        result = self.workflow.create_cpg(
            'test_cpg',
            domain='test_domain',
            growthSizeMiB=5120,
            ha='HAJBOD_JBOD'
        )
        
        posted = self.session_mgr.rest_client.posted_data[-1]
        payload = posted['payload']
        
        self.assertEqual(payload['domain'], 'test_domain')
        self.assertEqual(payload['growthSizeMiB'], 5120)
        self.assertEqual(payload['ha'], 'HAJBOD_JBOD')
    
    # ===================================================================
    # CPG ALREADY EXISTS TESTS
    # ===================================================================
    
    def test_create_cpg_already_exists(self):
        """Test creating CPG that already exists raises error"""
        self._simulate_cpg_exists('existing_cpg')
        with self.assertRaises(Exception) as context:
            self.workflow.create_cpg('existing_cpg')
        self.assertIn("already exists", str(context.exception))
    
    # ===================================================================
    # DELETE CPG TESTS
    # ===================================================================
    
    def test_delete_cpg_success(self):
        """Test deleting existing CPG"""
        self._simulate_cpg_exists('test_cpg')
        result = self.workflow.delete_cpg('test_cpg')
        
        self.assertEqual(result['status'], 'deleted')
        self.assertIn('cpg-test_cpg-uid', self.session_mgr.rest_client.deleted_endpoints[0])
    
    def test_delete_cpg_not_exists(self):
        """Test deleting non-existent CPG raises error"""
        with self.assertRaises(Exception) as context:
            self.workflow.delete_cpg('nonexistent_cpg')
        self.assertIn("does not exist", str(context.exception).lower())
    
    # ===================================================================
    # VALIDATOR TESTS - NAME VALIDATION
    # ===================================================================
    
    def test_create_cpg_name_too_long(self):
        """Test CPG name longer than 31 characters"""
        long_name = 'a' * 32
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg(long_name)
        self.assertIn("31 characters", str(context.exception))
    
    def test_create_cpg_name_not_string(self):
        """Test CPG name as non-string type"""
        with self.assertRaises(ValueError):
            self.workflow.create_cpg(123)
    
    # ===================================================================
    # VALIDATOR TESTS - DOMAIN VALIDATION
    # ===================================================================
    
    def test_create_cpg_domain_not_string(self):
        """Test domain parameter as non-string"""
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg('test_cpg', domain=123)
        self.assertIn("domain must be a string", str(context.exception))
    
    def test_create_cpg_domain_empty_string(self):
        """Test domain parameter as empty string"""
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg('test_cpg', domain='')
        self.assertIn("domain name cannot be empty", str(context.exception))
    
    # ===================================================================
    # VALIDATOR TESTS - CAGE VALIDATION
    # ===================================================================
    
    def test_create_cpg_cage_not_string(self):
        """Test cage parameter as non-string"""
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg('test_cpg', cage=123)
        self.assertIn("cage must be a string", str(context.exception))
    
    def test_create_cpg_cage_empty_string(self):
        """Test cage parameter as empty string"""
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg('test_cpg', cage='')
        self.assertIn("cage cannot be empty", str(context.exception))
    
    # ===================================================================
    # VALIDATOR TESTS - POSITION VALIDATION
    # ===================================================================
    
    def test_create_cpg_position_not_string(self):
        """Test position parameter as non-string"""
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg('test_cpg', position=123)
        self.assertIn("position must be a string", str(context.exception))
    
    def test_create_cpg_position_empty_string(self):
        """Test position parameter as empty string"""
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg('test_cpg', position='')
        self.assertIn("position cannot be empty", str(context.exception))
    
    # ===================================================================
    # VALIDATOR TESTS - GROWTH SIZE VALIDATION
    # ===================================================================
    
    def test_create_cpg_growthSizeMiB_not_number(self):
        """Test growthSizeMiB as non-numeric"""
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg('test_cpg', growthSizeMiB='invalid')
        self.assertIn("growthSizeMiB must be a number", str(context.exception))
    
    def test_create_cpg_growthSizeMiB_negative(self):
        """Test growthSizeMiB as negative value"""
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg('test_cpg', growthSizeMiB=-100)
        self.assertIn("growthSizeMiB must be a non-negative", str(context.exception))
    
    def test_create_cpg_growthSizeMiB_zero(self):
        """Test growthSizeMiB as zero (valid)"""
        result = self.workflow.create_cpg('test_cpg', growthSizeMiB=0)
        posted = self.session_mgr.rest_client.posted_data[-1]
        self.assertEqual(posted['payload']['growthSizeMiB'], 0)
    
    def test_create_cpg_growthSizeMiB_float(self):
        """Test growthSizeMiB as float value"""
        result = self.workflow.create_cpg('test_cpg', growthSizeMiB=5120.5)
        posted = self.session_mgr.rest_client.posted_data[-1]
        self.assertEqual(posted['payload']['growthSizeMiB'], 5120.5)
    
    # ===================================================================
    # VALIDATOR TESTS - GROWTH LIMIT VALIDATION
    # ===================================================================
    
    def test_create_cpg_growthLimitMiB_not_number(self):
        """Test growthLimitMiB as non-numeric"""
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg('test_cpg', growthLimitMiB='invalid')
        self.assertIn("growthLimitMiB must be a number", str(context.exception))
    
    def test_create_cpg_growthLimitMiB_negative(self):
        """Test growthLimitMiB as negative value"""
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg('test_cpg', growthLimitMiB=-100)
        self.assertIn("growthLimitMiB must be a non-negative", str(context.exception))
    
    def test_create_cpg_growthLimitMiB_zero(self):
        """Test growthLimitMiB as zero (valid)"""
        result = self.workflow.create_cpg('test_cpg', growthLimitMiB=0)
        posted = self.session_mgr.rest_client.posted_data[-1]
        self.assertEqual(posted['payload']['growthLimitMiB'], 0)
    
    # ===================================================================
    # VALIDATOR TESTS - GROWTH WARNING VALIDATION
    # ===================================================================
    
    def test_create_cpg_growthWarnMiB_not_number(self):
        """Test growthWarnMiB as non-numeric"""
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg('test_cpg', growthWarnMiB='invalid')
        self.assertIn("growthWarnMiB must be a number", str(context.exception))
    
    def test_create_cpg_growthWarnMiB_negative(self):
        """Test growthWarnMiB as negative value"""
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg('test_cpg', growthWarnMiB=-100)
        self.assertIn("growthWarnMiB must be a non-negative", str(context.exception))
    
    def test_create_cpg_growthWarnMiB_zero(self):
        """Test growthWarnMiB as zero (valid)"""
        result = self.workflow.create_cpg('test_cpg', growthWarnMiB=0)
        posted = self.session_mgr.rest_client.posted_data[-1]
        self.assertEqual(posted['payload']['growthWarnMiB'], 0)
    
    # ===================================================================
    # VALIDATOR TESTS - HA VALIDATION
    # ===================================================================
    
    def test_create_cpg_ha_not_string(self):
        """Test ha parameter as non-string"""
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg('test_cpg', ha=123)
        self.assertIn("ha must be a string", str(context.exception))
    
    def test_create_cpg_ha_invalid_value(self):
        """Test ha parameter with invalid value"""
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg('test_cpg', ha='INVALID_HA')
        self.assertIn("ha must be one of", str(context.exception))
    
    def test_create_cpg_ha_valid_HAJBOD_DISK(self):
        """Test ha parameter with valid HAJBOD_DISK value"""
        result = self.workflow.create_cpg('test_cpg', ha='HAJBOD_DISK')
        posted = self.session_mgr.rest_client.posted_data[-1]
        self.assertEqual(posted['payload']['ha'], 'HAJBOD_DISK')
    
    # ===================================================================
    # VALIDATOR TESTS - KEY VALUE PAIRS VALIDATION
    # ===================================================================
    
    def test_create_cpg_keyValuePairs_not_dict(self):
        """Test keyValuePairs as non-dict"""
        with self.assertRaises(ValueError) as context:
            self.workflow.create_cpg('test_cpg', keyValuePairs='invalid')
        self.assertIn("keyValuePairs must be a dictionary", str(context.exception))
    
    def test_create_cpg_keyValuePairs_empty_dict(self):
        """Test keyValuePairs as empty dict (valid)"""
        result = self.workflow.create_cpg('test_cpg', keyValuePairs={})
        posted = self.session_mgr.rest_client.posted_data[-1]
        self.assertEqual(posted['payload']['keyValuePairs'], {})
    
    # ===================================================================
    # PREPROCESSING TESTS - UNIT CONVERSION
    # ===================================================================
    
    def test_preprocess_growth_increment_with_unit(self):
        """Test preprocessing growth_increment with unit conversion"""
        params = {
            'growth_increment': 10,
            'growth_increment_unit': 'GiB'
        }
        result = self.workflow._preprocess_create_cpg(params)
        self.assertIn('growthSizeMiB', result)
        self.assertNotIn('growth_increment', result)
        self.assertNotIn('growth_increment_unit', result)
        self.assertEqual(result['growthSizeMiB'], 10240)  # 10 GiB = 10240 MiB
    
    def test_preprocess_growth_limit_with_unit(self):
        """Test preprocessing growth_limit with unit conversion"""
        params = {
            'growth_limit': 100,
            'growth_limit_unit': 'GiB'
        }
        result = self.workflow._preprocess_create_cpg(params)
        self.assertIn('growthLimitMiB', result)
        self.assertNotIn('growth_limit', result)
        self.assertNotIn('growth_limit_unit', result)
        self.assertEqual(result['growthLimitMiB'], 102400)  # 100 GiB = 102400 MiB
    
    def test_preprocess_growth_warning_with_unit(self):
        """Test preprocessing growth_warning with unit conversion"""
        params = {
            'growth_warning': 80,
            'growth_warning_unit': 'GiB'
        }
        result = self.workflow._preprocess_create_cpg(params)
        self.assertIn('growthWarnMiB', result)
        self.assertNotIn('growth_warning', result)
        self.assertNotIn('growth_warning_unit', result)
        self.assertEqual(result['growthWarnMiB'], 81920)  # 80 GiB = 81920 MiB
    
    def test_preprocess_none_params(self):
        """Test preprocessing with None params"""
        result = self.workflow._preprocess_create_cpg(None)
        self.assertEqual(result, {})
    
    def test_preprocess_mixed_unit_and_direct_params(self):
        """Test preprocessing with both unit conversion and direct MiB params"""
        params = {
            'growth_increment': 5,
            'growth_increment_unit': 'GiB',
            'growthLimitMiB': 102400,
            'domain': 'production'
        }
        result = self.workflow._preprocess_create_cpg(params)
        self.assertEqual(result['growthSizeMiB'], 5120)
        self.assertEqual(result['growthLimitMiB'], 102400)
        self.assertEqual(result['domain'], 'production')
    
    def test_preprocess_growth_increment_none_value(self):
        """Test preprocessing when growth_increment is None"""
        params = {
            'growth_increment': None,
            'growth_increment_unit': 'GiB'
        }
        result = self.workflow._preprocess_create_cpg(params)
        self.assertNotIn('growthSizeMiB', result)
    
    # ===================================================================
    # GET CPG INFO TESTS
    # ===================================================================
    
    def test_get_cpg_info_exists(self):
        """Test getting info for existing CPG"""
        self._simulate_cpg_exists('test_cpg')
        result = self.workflow.get_cpg_info('test_cpg')
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]['name'], 'test_cpg')
    
    def test_get_cpg_info_not_exists(self):
        """Test getting info for non-existent CPG returns empty list"""
        result = self.workflow.get_cpg_info('nonexistent')
        self.assertEqual(result, [])
    
    # ===================================================================
    # LIST CPGS TESTS
    # ===================================================================
    
    def test_list_cpgs_returns_members(self):
        """Test list_cpgs returns proper structure"""
        result = self.workflow.list_cpgs()
        self.assertIn('members', result)
        self.assertIsInstance(result['members'], list)
    
    # ===================================================================
    # EDGE CASES
    # ===================================================================
    
    def test_create_cpg_with_none_optional_params(self):
        """Test creating CPG with None values for optional params"""
        result = self.workflow.create_cpg('test_cpg', domain=None, growthSizeMiB=None)
        # None values should be included but validation should pass
        self.assertEqual(result['status'], 'created')
    
    def test_create_cpg_validation_params_not_dict(self):
        """Test validation with non-dict params raises error"""
        with self.assertRaises(ValueError) as context:
            validate_cpg_params('test', 'not_a_dict')
        self.assertIn("params must be a dictionary", str(context.exception))


class TestCpgWorkflowIntegration(unittest.TestCase):
    """Integration tests for CPG workflow"""
    
    def setUp(self):
        """Set up test fixtures"""
        self.session_mgr = MockSessionManager()
        self.task_mgr = MockTaskManager(self.session_mgr)
        self.workflow = CpgWorkflow(self.session_mgr, self.task_mgr)
    
    def _simulate_cpg_exists(self, name):
        """Helper to mark a CPG as existing"""
        self.session_mgr.rest_client.existing_cpgs.add(name)
    
    def test_create_and_delete_cpg_flow(self):
        """Test complete create and delete flow"""
        # Create CPG
        create_result = self.workflow.create_cpg('integration_cpg')
        self.assertEqual(create_result['status'], 'created')
        
        # Verify CPG was marked as existing by create operation
        self.assertIn('integration_cpg', self.session_mgr.rest_client.existing_cpgs)
        
        # Get CPG info
        info_result = self.workflow.get_cpg_info('integration_cpg')
        self.assertEqual(info_result[0]['name'], 'integration_cpg')
        
        # Delete CPG
        delete_result = self.workflow.delete_cpg('integration_cpg')
        self.assertEqual(delete_result['status'], 'deleted')
    
    def test_create_cpg_with_complex_configuration(self):
        """Test creating CPG with complex real-world configuration"""
        result = self.workflow.create_cpg(
            'prod_cpg',
            domain='production',
            growthSizeMiB=5120,
            growthLimitMiB=512000,
            growthWarnMiB=409600,
            ha='HAJBOD_JBOD',
            cage='cage1',
            keyValuePairs={'env': 'prod', 'team': 'storage'}
        )
        
        posted = self.session_mgr.rest_client.posted_data[-1]
        payload = posted['payload']
        
        # Verify all parameters
        self.assertEqual(payload['name'], 'prod_cpg')
        self.assertEqual(payload['domain'], 'production')
        self.assertEqual(payload['growthSizeMiB'], 5120)
        self.assertEqual(payload['growthLimitMiB'], 512000)
        self.assertEqual(payload['growthWarnMiB'], 409600)
        self.assertEqual(payload['ha'], 'HAJBOD_JBOD')
        self.assertEqual(payload['cage'], 'cage1')
        self.assertEqual(payload['keyValuePairs']['env'], 'prod')
        self.assertEqual(payload['keyValuePairs']['team'], 'storage')
    
    def test_delete_cpg_missing_uid(self):
        """Test delete CPG when UID is missing from response"""
        # Mock get_cpg_info to return CPG without UID
        original_get = self.workflow.get_cpg_info
        self.workflow.get_cpg_info = lambda name: [{'name': name}]
        
        with self.assertRaises(ValueError) as context:
            self.workflow.delete_cpg('test_cpg')
        
        self.assertIn('UID missing', str(context.exception))
        
        # Restore
        self.workflow.get_cpg_info = original_get
    
    def test_get_cpg_info_hpe_storage_exception(self):
        """Test get_cpg_info when HPEStorageException is raised"""
        # Mock rest_client.get to raise HPEStorageException
        original_get = self.session_mgr.rest_client.get
        
        def mock_get_exception(endpoint, headers=None):
            raise HPEStorageException("Storage subsystem error")
        
        self.session_mgr.rest_client.get = mock_get_exception
        
        with self.assertRaises(HPEStorageException):
            self.workflow.get_cpg_info('test_cpg')
        
        # Restore
        self.session_mgr.rest_client.get = original_get
    
    def test_list_cpgs_hpe_storage_exception(self):
        """Test list_cpgs when HPEStorageException is raised"""
        # Mock rest_client.get to raise HPEStorageException
        original_get = self.session_mgr.rest_client.get
        
        def mock_get_exception(endpoint, headers=None):
            raise HPEStorageException("Storage subsystem error")
        
        self.session_mgr.rest_client.get = mock_get_exception
        
        with self.assertRaises(HPEStorageException):
            self.workflow.list_cpgs()
        
        # Restore
        self.session_mgr.rest_client.get = original_get
    
    def test_create_cpg_rest_client_exception(self):
        """Test create_cpg when REST client post raises exception"""
        # Mock rest_client.post to raise an exception
        original_post = self.session_mgr.rest_client.post
        
        def mock_post_exception(endpoint, payload):
            raise RuntimeError("Network error")
        
        self.session_mgr.rest_client.post = mock_post_exception
        
        with self.assertRaises(Exception):
            self.workflow.create_cpg('test_cpg')
        
        # Restore
        self.session_mgr.rest_client.post = original_post
    
    def test_delete_cpg_rest_client_exception(self):
        """Test delete_cpg when REST client delete raises exception"""
        # First ensure the CPG exists
        self.session_mgr.rest_client.existing_cpgs.add('test_cpg')
        
        # Mock rest_client.delete to raise an exception
        original_delete = self.session_mgr.rest_client.delete
        
        def mock_delete_exception(endpoint):
            raise RuntimeError("Network error")
        
        self.session_mgr.rest_client.delete = mock_delete_exception
        
        with self.assertRaises(Exception):
            self.workflow.delete_cpg('test_cpg')
        
        # Restore
        self.session_mgr.rest_client.delete = original_delete


if __name__ == '__main__':
    unittest.main()



#    (c) Copyright 2026 Hewlett Packard Enterprise Development LP
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
def validate_qos_params(name=None, maxIOPS=None, maxBWS=None, enable=None, **kwargs):
    if name is not None:
        if not isinstance(name, str) or not name.strip():
            raise ValueError("QOS name must be a non-empty string.")

    if maxIOPS is not None:
        if not isinstance(maxIOPS, int) or maxIOPS <= 0:
            raise ValueError("maxIOPS must be a positive integer.")

    if maxBWS is not None:
        if not isinstance(maxBWS, int) or maxBWS <= 0:
            raise ValueError("maxBWS must be a positive integer.")

    if enable is not None:
        if not isinstance(enable, bool):
            raise ValueError("enable must be a boolean.")


#    (c) Copyright 2026 Hewlett Packard Enterprise Development LP
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import requests
from hpe_storage_flowkit_py.v3.src.core.exceptions import HPEStorageException, HTTPError
from hpe_storage_flowkit_py.v3.src.core.logger import Logger
log=Logger()
class RESTClient:
    # Sensitive fields that should be masked in logs
    SENSITIVE_FIELDS = {'password', 'token', 'key', 'secret', 'authorization'}
    
    def __init__(self, api_url):
        self.api_url = api_url.rstrip('/')
        self.session = requests.Session()
        self.verify_ssl = False

    def _mask_sensitive_data(self, data):
        """Recursively mask sensitive fields in dictionaries and lists for logging."""
        if isinstance(data, dict):
            masked_data = {}
            for key, value in data.items():
                if key.lower() in self.SENSITIVE_FIELDS:
                    masked_data[key] = '********'
                else:
                    masked_data[key] = self._mask_sensitive_data(value)
            return masked_data
        elif isinstance(data, list):
            return [self._mask_sensitive_data(item) for item in data]
        else:
            return data

    def _make_url(self, endpoint):
        if not endpoint.startswith('/'):
            endpoint = '/' + endpoint
        return f"{self.api_url}{endpoint}"

    def get(self, endpoint, **kwargs):
        url = self._make_url(endpoint)
        log.debug(f"GET request to: {url}")
        response = self.session.get(url, verify=self.verify_ssl, **kwargs)
        self._check_response(response)
        log.debug(f"GET response status: {response.status_code}")
        parsed_response = self._parse_response(response)
        log.debug(f"GET response data: {self._mask_sensitive_data(parsed_response)}")
        return parsed_response

    def post(self, endpoint, payload=None, **kwargs):
        url = self._make_url(endpoint)
        log.debug(f"POST request to: {url}")
        log.debug(f"POST payload: {self._mask_sensitive_data(payload)}")
        response = self.session.post(url, json=payload, verify=self.verify_ssl, **kwargs)
        self._check_response(response)
        log.debug(f"POST response status: {response.status_code}")
        parsed_response = self._parse_response(response)
        log.debug(f"POST response data: {self._mask_sensitive_data(parsed_response)}")
        return parsed_response

    def put(self, endpoint, payload=None, **kwargs):
        url = self._make_url(endpoint)
        log.debug(f"PUT request to: {url}")
        log.debug(f"PUT payload: {payload}")
        response = self.session.put(url, json=payload, verify=self.verify_ssl, **kwargs)
        self._check_response(response)
        log.debug(f"PUT response status: {response.status_code}")
        parsed_response = self._parse_response(response)
        log.debug(f"PUT response data: {self._mask_sensitive_data(parsed_response)}")
        return parsed_response
    def patch(self, endpoint, payload=None, **kwargs):
        url = self._make_url(endpoint)
        log.debug(f"PATCH request to: {url}")
        log.debug(f"PATCH payload: {payload}")
        response = self.session.patch(url, json=payload, verify=self.verify_ssl, **kwargs)
        self._check_response(response)
        log.debug(f"PATCH response status: {response.status_code}")
        parsed_response = self._parse_response(response)
        log.debug(f"PATCH response data: {self._mask_sensitive_data(parsed_response)}")
        return parsed_response
    def delete(self, endpoint, **kwargs):
        url = self._make_url(endpoint)
        log.debug(f"DELETE request to: {url}")
        response = self.session.delete(url, verify=self.verify_ssl, **kwargs)
        self._check_response(response)
        log.debug(f"DELETE response status: {response.status_code}")
        parsed_response = self._parse_response(response)
        log.debug(f"DELETE response data: {self._mask_sensitive_data(parsed_response)}")
        return parsed_response

    def _check_response(self, response):
        if not response.ok:
            log.error(f"HTTP {response.status_code} error: {response.text}")
            raise HTTPError(response.status_code, response.text)
    def _parse_response(self, response):
        if response.text:
            try:
                return response.json()
            except Exception:
                return response.text
        return None



#    (c) Copyright 2026 Hewlett Packard Enterprise Development LP
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from urllib.parse import quote
from hpe_storage_flowkit_py.v1.src.core import exceptions
from hpe_storage_flowkit_py.v1.src.core.exceptions import HPEStorageException
from hpe_storage_flowkit_py.v1.src.core.session import SessionManager
from hpe_storage_flowkit_py.v1.src.validators.vlun_validator import validate_vlun_params

class VLUNWorkflow:
    """
    VLUNWorkflow
    """

    PORT_MODE_TARGET = 2
    PORT_MODE_INITIATOR = 3
    PORT_MODE_PEER = 4

    PORT_PROTO_FC = 1
    PORT_PROTO_ISCSI = 2
    PORT_PROTO_FCOE = 3
    PORT_PROTO_IP = 4
    PORT_PROTO_SAS = 5
    PORT_PROTO_NVME = 6

    PORT_STATE_READY = 4
    PORT_STATE_SYNC = 5
    PORT_STATE_OFFLINE = 10

    DEFAULT_NVME_PORT = 4420
    DEFAULT_PORT_NQN = 'nqn.2014-08.org.nvmexpress.discovery'

    def __init__(self, session_client: SessionManager):
        """
        __init__
        """
        self.session_client = session_client

    def export_volume_to_host(self, payload):
        """
        export_volume_to_host
        """
        return self.session_client.rest_client.post("/vluns", payload)

    def unexport_volume_from_host(self, vlun_id):
        """
        unexport_volume_from_host
        """
        return self.session_client.rest_client.delete(f"/vluns/{vlun_id}")

    def export_volume_to_hostset(self, payload):
        """
        export_volume_to_hostset
        """
        return self.session_client.rest_client.post("/vluns", payload)

    def unexport_volume_from_hostset(self, vlun_id):
        """
        unexport_volume_from_hostset
        """
        return self.session_client.rest_client.delete(f"/vluns/{vlun_id}")

    def export_volumeset_to_host(self, payload):
        """
        export_volumeset_to_host
        """
        return self.session_client.rest_client.post("/vluns", payload)

    def unexport_volumeset_from_host(self, vlun_id):
        """
        unexport_volumeset_from_host
        """
        return self.session_client.rest_client.delete(f"/vluns/{vlun_id}")

    def export_volumeset_to_hostset(self, payload):
        """
        export_volumeset_to_hostset
        """
        return self.session_client.rest_client.post("/vluns", payload)

    def unexport_volumeset_from_hostset(self, vlun_id):
        """
        unexport_volumeset_from_hostset
        """
        return self.session_client.rest_client.delete(f"/vluns/{vlun_id}")

    def vlun_exists(self, volume_name, lun, host_name=None, port_pos=None):
        """
        vlun_exists
        """
        vluns = self.list_vluns()
        for vlun in vluns:
            if (
                vlun.get("volumeName") == volume_name
                and vlun.get("lun") == lun
                and (host_name is None or vlun.get("hostname") == host_name)
                and (port_pos is None or vlun.get("portPos") == port_pos)
            ):
                return True
        return False

    def list_vluns(self):
        """
        list_vluns
        """
        response = self.session_client.rest_client.get("/vluns")
        if response is None:
            return []
        return response.get("members", [])

    def get_vvsets(self, volume_set_name):
        """
        get_vvsets
        """
        return self.session_client.rest_client.get(f"/volumesets/{volume_set_name}")


    # below functions reqd for cinder

    def create_vlun(self, volume_name, host_name, params):
        validate_vlun_params(volume_name, host_name, params)
        payload = {"volumeName": volume_name, "hostname": host_name}
        payload.update(params)
        try:
            response = self.session_client.rest_client.post("/vluns", payload)
            return response
        except HPEStorageException as e:
            raise

    def delete_vlun(self, volumeName, lunID, hostname=None, port=None):
        #validate_vlun_params(lunID)

        vlun = "%s,%s" % (volumeName, lunID)

        if hostname:
            vlun += ",%s" % hostname
        else:
            if port:
                vlun += ","

        if port:
            vlun += ",%s:%s:%s" % (port['node'],
                                   port['slot'],
                                   port['cardPort'])

        #print("vlun: ", vlun)
        try:
            #response = self.session_client.rest_client.delete(f"/vluns/{vlun_id}")
            response = self.session_client.rest_client.delete(f"/vluns/{vlun}")
            return response
        except exceptions.HTTPNotFound as e:
            raise
        except HPEStorageException as e:
            raise

    def get_vlun(self, vlun_id):
        validate_vlun_params(vlun_id)
        try:
            response = self.session_client.rest_client.get(f"/vluns/{vlun_id}")
            return response
        except HPEStorageException as e:
            raise

    def getHost(self, name):
        """Get information about a Host.

        :param name: The name of the Host to find
        :type name: str

        :returns: host dict
        :raises: :class:`~hpe3parclient.exceptions.HTTPNotFound`
            - NON_EXISTENT_HOST - HOST doesn't exist

        """
        response = self.session_client.rest_client.get(f"/hosts/{name}")
        return response

    def getHostVLUNs(self, hostName):
        """Get all of the VLUNs on a specific Host.

        :param hostName: Host name
        :type hostNane: str

        :returns: list of vluns
        :raises: :class:`~hpe3parclient.exceptions.HTTPNotFound`
            - NON_EXISTENT_HOST - HOST doesn't exist

        """
        # calling getHost to see if the host exists and raise not found
        # exception if it's not found.
        self.getHost(hostName)

        vluns = []

        #response, body = self.http.get('/vluns?query=%s' %
        #                               quote(query.encode("utf8")))
        uri = '/vluns?query="hostname EQ %s"' % hostName
        response = self.session_client.rest_client.get(uri)

        for vlun in response.get('members', []):
            vluns.append(vlun)

        return vluns

    def getVLUNs(self):
        """Get all VLUNs.
        In case of 'force detach', hostname is None. Also vlun_id is not known.
        In such scenario, we need list of all VLUNs.

        :returns: Array of VLUNs

        """
        try:
            response = self.session_client.rest_client.get(f"/vluns")
            return response
        except HPEStorageException as e:
            raise

    # PORT Methods
    def _getIscsiVlan(self, nsp):
        """Get iSCSI VLANs for an iSCSI port

        :param nsp: node slot port Eg. '0:2:1'
        :type nsp: str

        :returns: list of iSCSI VLANs

        """
        #response, body = self.http.get('/ports/' + nsp + '/iSCSIVlans/')
        response = self.session_client.rest_client.get(f"/ports/{nsp}/iSCSIVlans/")
        return response

    def getPorts(self):
        """Get the list of ports on the 3PAR.

        :returns: list of Ports

        """
        #response, body = self.http.get('/ports')
        body = self.session_client.rest_client.get(f"/ports")

        # if any of the ports are iSCSI ports and
        # are vlan tagged (as obtained by _getIscsiVlan), then
        # the vlan information is merged with the
        # returned port information.
        for port in body['members']:
            if (port['protocol'] == 2 and
                    'iSCSIPortInfo' in port and
                    port['iSCSIPortInfo']['vlan'] == 1):

                portPos = port['portPos']
                nsp_array = [str(portPos['node']), str(portPos['slot']),
                             str(portPos['cardPort'])]
                nsp = ":".join(nsp_array)
                vlan_body = self._getIscsiVlan(nsp)
                if vlan_body:
                    port['iSCSIVlans'] = vlan_body['iSCSIVlans']

        return body

    def getNvmePorts(self):
        all_ports = self.getPorts()

        target_ports = []
        for port in all_ports['members']:
            if (
                port['mode'] == self.PORT_MODE_TARGET and
                port['linkState'] == self.PORT_STATE_READY
            ):
                port_pos = port['portPos']
                nsp = '%s:%s:%s' % (port_pos['node'], port_pos['slot'],
                                    port_pos['cardPort'])
                port['nsp'] = nsp
                target_ports.append(port)

        nvme_ports = []
        for port in target_ports:
            if port['protocol'] == self.PORT_PROTO_NVME:
                nvme_ports.append(port)

        #logger.debug("nvme_ports: %(ports)s", {'ports': nvme_ports})
        print("nvme_ports: %(ports)s", {'ports': nvme_ports})
        return nvme_ports 

    def get_matched_array_ips_and_ports(self, client_conf):
        temp_nvme_ip = {}
        nvme_ip_list = {}
        conf_ips = client_conf['hpe3par_nvme_ips']

        for ip_addr in conf_ips:
            # "ip"(given by user in cinder conf)
            # contains IP Address, NVMe port in <IP>:<PORT> format.
            ip = ip_addr.split(':')
            if len(ip) == 1:
                # "ip" doesn't contain NVMe port, use default NVMe port.
                temp_nvme_ip[ip_addr] = {'ip_port': self.DEFAULT_NVME_PORT}
            elif len(ip) == 2:
                #Valid format <IP>:<PORT>
                temp_nvme_ip[ip[0]] = {'ip_port': ip[1]}
            else:
                #Invalid format such as <IP>:<PORT>:<DATA>
                #logger.warning("Invalid IP address format '%s'", ip_addr)
                print("Invalid IP address format '%s'", ip_addr)

        # get all the valid nvme ports from array
        # when found, add the valid nvme ip and port
        # to the nvme IP dictionary
        nvme_ports = self.getNvmePorts()
        if not len(nvme_ports):
            msg = 'Unable to obtain NVMe ports from storage system.'
            #logger.error(msg)
            #raise exceptions.HTTPNotFound(
            #    {'code': 'NON_EXISTENT_NVME_PORTS',
            #     'desc': msg})
            print("error: ", msg)
            raise Exception(msg)
        for port in nvme_ports:
            ip = port['nodeWWN']
            if ip in temp_nvme_ip:
                ip_port = temp_nvme_ip[ip]['ip_port']
                nvme_ip_list[ip] = {'ip_port': ip_port,
                                    'nsp': port['nsp']}
                del temp_nvme_ip[ip]

        #logger.debug("After mapping IPs from cinder to storage system:"
        #              "temp_nvme_ip: %(ips)s", {'ips': temp_nvme_ip})
        #logger.debug("After mapping IPs from cinder to storage system:"
        #             "nvme_ip_list: %(ips)s", {'ips': nvme_ip_list})
        print("After mapping IPs from cinder to storage system:"
                      "temp_nvme_ip: %(ips)s", {'ips': temp_nvme_ip})
        print("After mapping IPs from cinder to storage system:"
                     "nvme_ip_list: %(ips)s", {'ips': nvme_ip_list})

        # lets see if there are invalid nvme IPs left in the temp dict
        if len(temp_nvme_ip) > 0:
            #logger.warning("Found invalid nvme IP address(s) in "
            #            "configuration option(s) hpe3par_nvme_ips '%s.'",
            #            (", ".join(temp_nvme_ip)))
            warn_msg = "Found invalid nvme IP address(s) in "
            warn_msg = warn_msg + "configuration option(s) hpe3par_nvme_ips '%s.', ".join(temp_nvme_ip)
            print(warn_msg)

        if not len(nvme_ip_list):
            msg = 'At least one valid nvme IP address must be set.'
            #logger.error(msg)
            #raise exceptions.HTTPNotFound(
            #    {'code': 'NON_EXISTENT_NVME_IP',
            #     'desc': msg})
            print("error: ", msg)
            raise Exception(msg)

        ret_vals = (nvme_ip_list, nvme_ports)
        return ret_vals

    def getHosts(self):
        """Get information about every Host on the 3Par array.

        :returns: list of Hosts
        """
        response = self.session_client.rest_client.get(f"/hosts")
        return response

    def getHostByNqn(self, nqn):
        """Get information about a Host by its NVMe Qualified Name (NQN).

        :param nqn: The NQN of the Host to find
        :type nqn: str

        :returns: host dict
        :raises: HTTPNotFound
            - NON_EXISTENT_HOST - HOST doesn't exist

        """
        body = self.getHosts()
        if 'members' not in body:
            return None

        for host in body['members']:
            if 'NVMETCPPaths' not in host or not host['NVMETCPPaths']:
                continue
            nvme_paths = host['NVMETCPPaths']
            for path in nvme_paths:
                path_nqn = path.get('NQN')
                if path_nqn == nqn:
                    return host

        # If we reach here, no host with the given NQN was found
        #logger.error("Host with NQN '%s' not found.", nqn)
        msg = "Host with NQN '%s' not found." % nqn
        #raise exceptions.HTTPNotFound(error={'desc': msg})
        print("Error: Host with NQN '%s' not found.", nqn)
        msg = "Host with NQN '%s' not found." % nqn
        #raise Exception(msg)
        return None

    def getNqn(self, portPos=None):
        # in dev and QA array, all ports have same nqn below:
        # 'nqn.2014-08.org.nvmexpress.discovery'
        return self.DEFAULT_PORT_NQN

    def build_portPos(self, nsp):
        arr = nsp.split(":")
        portPos = {}
        portPos['node'] = int(arr[0])
        portPos['slot'] = int(arr[1])
        portPos['cardPort'] = int(arr[2])
        return portPos

    def find_existing_vluns(self, vol_name_3par, host):
        existing_vluns = []
        try:
            host_vluns = self.getHostVLUNs(host['name'])
            for vlun in host_vluns:
                if vlun['volumeName'] == vol_name_3par:
                    existing_vluns.append(vlun)
        #except exceptions.HTTPNotFound:
        except Exception as ex:
            # ignore, no existing VLUNs were found
            print("No existing VLUNs were found for host/volume "
                         "combination: %(host)s, %(vol)s",
                         {'host': host['name'],
                          'vol': vol_name_3par})
        return existing_vluns

    def getVLUN(self, volumeName, allVluns=False):
        """Get information about a VLUN.

        :raises: :class:`~hpe3parclient.exceptions.HTTPNotFound`
            -  NON_EXISTENT_VLUN - VLUN doesn't exist

        """
        #query = '"volumeName EQ %s"' % volumeName
        #response, body = self.http.get('/vluns?query=%s' %
        #                               quote(query.encode("utf8")))
        uri = '/vluns?query="volumeName EQ %s"' % volumeName
        response = self.session_client.rest_client.get(uri)

        if allVluns:
            # Return all the VLUNs found for the volume.
            vluns = response.get('members', [])
            return vluns
        else:
            # Return the first VLUN found for the volume.
            for vlun in response.get('members', []):
                return vlun

    def getNextLunId(self, hostname, host_type='nvme'):
        # lun id can be 0 through 16383 (1 to 256 for NVMe hosts)
        LIMIT = 16383
        if host_type=='nvme':
            LIMIT = 256

        lun_id_max = 0
        try:
            host_vluns = self.getHostVLUNs(hostname)
            for vlun in host_vluns:
                lun_id_x = vlun['lun']
                if lun_id_x > lun_id_max:
                    lun_id_max = lun_id_x

        #except exceptions.HTTPNotFound:
        except HPEStorageException as e:
            # ignore, no existing VLUNs were found
            pass

        lun_id_next = lun_id_max + 1
        if lun_id_next > LIMIT:
            msg = "Lun id exceeded limit '%d'" % LIMIT
            #raise exceptions.HTTPNotFound(error={'desc': msg})
            raise Exception(msg)

        return lun_id_next

    def create_vlun_nvme(self, vol_name_3par, host, nvme_ips):
        """Create a VLUN for NVMe host.
        :param vol_name_3par: The name of the volume on 3PAR.
        :param host: The host object containing host information.
        :param nvme_ips: The NVMe IPs to use for the VLUN.
        :returns: A tuple containing a list of portals and target NQNs.
        :rtype: tuple
        """

        # Collect all existing VLUNs for this volume/host combination.
        existing_vluns = self.find_existing_vluns(vol_name_3par, host)
        #logger.debug("existing_vluns: %(ev)s", {'ev': existing_vluns})
        print("existing_vluns: %(ev)s", {'ev': existing_vluns})
        host_name = host['name']
        portals = []
        target_nqns = []
        lun_id = None
        # check for an already existing VLUN matching the
        # nsp for this nvme IP. If one is found, use it
        # instead of creating a new VLUN.
        if existing_vluns:
            for v in existing_vluns:
                lun_id = v['lun']
                #logger.debug("vlun exists for host name: %(host)s" \
                #             " with lun: %(lun)s",
                #             {'host': host_name, 'lun': v['lun']})
                print("vlun exists for host name: %(host)s" \
                        " with lun: %(lun)s",
                        {'host': host_name, 'lun': v['lun']})
                break 
        else:
            #logger.debug("creating vlun for host name: %(host)s",
            #             {'host': host_name})
            print("creating vlun for host name: %(host)s",
                    {'host': host_name})
            if lun_id is None:
                #logger.debug("lun_id is None. calling getNextLunId")
                lun_id = self.getNextLunId(host_name)
                #logger.debug("lun_id_next: %(id)s", {'id': lun_id})

        #logger.debug("lun_id is %(id)s", {'id': lun_id})
        #location = self.createVLUN(vol_name_3par, lun=lun_id,
        #                           hostname=host_name)
        #logger.debug("location: %(loc)s", {'loc': location})
        params = {}
        params['lun'] = lun_id
        self.create_vlun(vol_name_3par, host_name, params)
        target_portal_ips = list(nvme_ips.keys())
        for nvme_ip in target_portal_ips:
            portals.append(
                (nvme_ip, nvme_ips[nvme_ip]['ip_port'], 'tcp')
                )
        vlun = self.getVLUN(vol_name_3par)
        nqn_of_vlun = vlun['Subsystem_NQN']
        #logger.debug("nqn_of_vlun: %(nqn)s", {'nqn': nqn_of_vlun})
        target_nqns.append(nqn_of_vlun)

        ret_vals = (portals, target_nqns)
        return ret_vals

    def remove_vlun_nvme(self, vol_name_3par, hostname, host_nqn):
        vlunsData = self.getVLUN(vol_name_3par, True)
        print("vlunsData: ", vlunsData)
        if vlunsData == []:
            #logger.error("No VLUN found for volume %(name)s on host %(host)s",
            #             {'name': vol_name_3par, 'host': hostname})
            print("Error: No VLUN found for volume %(name)s on host %(host)s",
                         {'name': vol_name_3par, 'host': hostname})
            return

        # When deleteing VLUNs, you simply need to remove the template VLUN
        # and any active VLUNs will be automatically removed.  The template
        # VLUN are marked as active: False
        vluns = []
        for vlun in vlunsData:
            if vol_name_3par in vlun['volumeName']:
                # template VLUNs are 'active' = False
                if not vlun['active']:
                    vluns.append(vlun)

        print("vluns: ", vluns)
        if not vluns:
            #logger.warning("3PAR vlun for volume %(name)s not found on host "
            #               "%(host)s", {'name': vol_name_3par, 'host': hostname})
            print("Warning: 3PAR vlun for volume %(name)s not found on host "
                        "%(host)s", {'name': vol_name_3par, 'host': hostname})
            return

        for vlun in vluns:
            print("  -- vlun: ", vlun)
            # Check if this VLUN belongs to the specified hostname
            if vlun.get('hostname') == hostname:
                #logger.debug("deleting vlun: %(lun)s", {'lun': vlun})
                print("deleting vlun")
                self.delete_vlun(vol_name_3par, vlun['lun'],
                                hostname)
            else:
                #logger.debug("Skipping vlun: %(lun)s -"
                print("Skipping vlun: %(lun)s -"
                             " belongs to different host: %(vlun_host)s",
                             {'lun': vlun['lun'],
                              'vlun_host': vlun.get('hostname', 'Unknown')})





#    (c) Copyright 2026 Hewlett Packard Enterprise Development LP
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import unittest
from unittest.mock import MagicMock, Mock, patch
import sys
import os

# Ensure src is on sys.path for direct imports
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from hpe_storage_flowkit_py.v3.src.workflows.hostset import HostSetWorkflow
from hpe_storage_flowkit_py.v3.src.core import exceptions
from hpe_storage_flowkit_py.v3.src.core.rest_client import RESTClient


class TestHostSetWorkflow(unittest.TestCase):
	"""Unit tests for HostSetWorkflow class using a simple Mock REST client.

	Tests cover all methods with positive and negative scenarios including edge cases.
	"""

	def setUp(self):
		"""Set up test fixtures before each test method."""
		self.session_client = Mock()
		self.session_client.rest_client = Mock(spec=RESTClient)
		# Default canned responses (no state tracking)
		self.session_client.rest_client.get.return_value = {"members": {}}
		self.session_client.rest_client.post.return_value = {"status": "created"}
		self.session_client.rest_client.delete.return_value = {"status": "deleted"}
		self.session_client.rest_client.patch.return_value = {"status": "modified"}
		self.task_manager = Mock()
		self.task_manager.waitForTaskToEnd.return_value = {"status": "completed"}
		self.workflow = HostSetWorkflow(self.session_client, self.task_manager)

	def _simulate_exists(self, name, uid="uid_x", members=None, domain=None, comment=None):
		"""Helper to simulate an existing host set with given properties."""
		members = members or []
		hostset_data = {
			"uid": uid,
			"name": name,
			"members": members
		}
		if domain:
			hostset_data["domain"] = domain
		if comment:
			hostset_data["comment"] = comment
		
		self.session_client.rest_client.get.return_value = {
			"members": {
				uid: hostset_data
			}
		}
	
	# ===================================================================
	# CREATE HOSTSET TESTS
	# ===================================================================
	
	def test_create_hostset_success_minimal(self):
		"""Test successful hostset creation with only required name parameter."""
		self.workflow.create_hostset("hostset1")
		self.session_client.rest_client.post.assert_called_once()
		call_args = self.session_client.rest_client.post.call_args
		self.assertEqual(call_args[0][0], "/hostsets")
		self.assertEqual(call_args[0][1]["name"], "hostset1")
	
	def test_create_hostset_success_with_all_optional_params(self):
		"""Test hostset creation with all optional parameters."""
		self.workflow.create_hostset(
			"hostset2",
			domain="production",
			comment="Production host set",
			setmembers=["host1", "host2", "host3"]
		)
		call_args = self.session_client.rest_client.post.call_args
		payload = call_args[0][1]  # Second positional argument
		self.assertEqual(payload["name"], "hostset2")
		self.assertEqual(payload["domain"], "production")
		self.assertEqual(payload["comment"], "Production host set")
		self.assertEqual(payload["members"], ["host1", "host2", "host3"])
	
	def test_create_hostset_success_with_domain_only(self):
		"""Test hostset creation with only domain parameter."""
		self.workflow.create_hostset("hostset3", domain="test_domain")
		call_args = self.session_client.rest_client.post.call_args
		payload = call_args[0][1]
		self.assertEqual(payload["name"], "hostset3")
		self.assertEqual(payload["domain"], "test_domain")
	
	def test_create_hostset_success_with_comment_only(self):
		"""Test hostset creation with only comment parameter."""
		self.workflow.create_hostset("hostset4", comment="Test comment")
		call_args = self.session_client.rest_client.post.call_args
		payload = call_args[0][1]
		self.assertEqual(payload["name"], "hostset4")
		self.assertEqual(payload["comment"], "Test comment")
	
	def test_create_hostset_success_with_members_only(self):
		"""Test hostset creation with only members parameter."""
		self.workflow.create_hostset("hostset5", setmembers=["host1", "host2"])
		call_args = self.session_client.rest_client.post.call_args
		payload = call_args[0][1]
		self.assertEqual(payload["name"], "hostset5")
		self.assertEqual(payload["members"], ["host1", "host2"])
	
	def test_create_hostset_async_task_202(self):
		"""Test hostset creation with async task (202 response)."""
		# Mock 202 response with resourceUri
		self.session_client.rest_client.post.return_value = {
			"message": "Started task to execute create Host Set",
			"resourceUri": "/api/v3/tasks/abc123"
		}
		# Mock task manager wait_for_task_to_end
		with patch.object(self.workflow.task_manager, 'wait_for_task_to_end', return_value={"status": "completed"}):
			self.workflow.create_hostset("hostset_async")
			self.workflow.task_manager.wait_for_task_to_end.assert_called_once_with(
				"/api/v3/tasks/abc123"
			)
	
	def test_create_hostset_invalid_name_empty(self):
		"""Test hostset creation with empty name."""
		with self.assertRaises(ValueError) as context:
			self.workflow.create_hostset("")
		self.assertIn("non-empty string", str(context.exception))
	
	def test_create_hostset_invalid_name_none(self):
		"""Test hostset creation with None name."""
		with self.assertRaises(ValueError) as context:
			self.workflow.create_hostset(None)
		self.assertIn("cannot be null", str(context.exception))
	
	def test_create_hostset_invalid_name_whitespace_only(self):
		"""Test hostset creation with whitespace-only name."""
		with self.assertRaises(ValueError) as context:
			self.workflow.create_hostset("   ")
		self.assertIn("non-empty string", str(context.exception))
	
	def test_create_hostset_invalid_name_too_long(self):
		"""Test hostset creation with name exceeding max length of 27 characters."""
		long_name = "a" * 28  # Exceeds NAME_MAX of 27
		with self.assertRaises(ValueError) as context:
			self.workflow.create_hostset(long_name)
		self.assertIn("27 characters", str(context.exception))
	
	def test_create_hostset_invalid_name_below_min(self):
		"""Test hostset creation with name less than min length."""
		# Empty string is handled by non-empty check, but test boundary
		with self.assertRaises(ValueError):
			self.workflow.create_hostset("")
	
	def test_create_hostset_invalid_domain_not_string(self):
		"""Test hostset creation with non-string domain."""
		with self.assertRaises(ValueError) as context:
			self.workflow.create_hostset("hostset6", domain=123)
		self.assertIn("must be a string", str(context.exception).lower())
	
	def test_create_hostset_invalid_domain_empty(self):
		"""Test hostset creation with empty domain string."""
		with self.assertRaises(ValueError) as context:
			self.workflow.create_hostset("hostset7", domain="")
		self.assertIn("cannot be empty", str(context.exception).lower())
	
	def test_create_hostset_invalid_domain_whitespace_only(self):
		"""Test hostset creation with whitespace-only domain."""
		with self.assertRaises(ValueError) as context:
			self.workflow.create_hostset("hostset8", domain="   ")
		self.assertIn("cannot be empty", str(context.exception).lower())
	
	def test_create_hostset_invalid_comment_not_string(self):
		"""Test hostset creation with non-string comment."""
		with self.assertRaises(ValueError) as context:
			self.workflow.create_hostset("hostset9", comment=456)
		self.assertIn("must be a string", str(context.exception).lower())
	
	def test_create_hostset_invalid_members_not_list(self):
		"""Test hostset creation with members not being a list."""
		with self.assertRaises(ValueError) as context:
			self.workflow.create_hostset("hostset10", setmembers="host1")
		self.assertIn("must be a list", str(context.exception).lower())
	
	def test_create_hostset_invalid_members_empty_string(self):
		"""Test hostset creation with empty string in members list."""
		with self.assertRaises(ValueError) as context:
			self.workflow.create_hostset("hostset11", setmembers=["host1", ""])
		self.assertIn("non-empty string", str(context.exception))
	
	def test_create_hostset_invalid_members_non_string_element(self):
		"""Test hostset creation with non-string element in members list."""
		with self.assertRaises(ValueError) as context:
			self.workflow.create_hostset("hostset12", setmembers=["host1", 123])
		self.assertIn("non-empty string", str(context.exception))
	
	def test_create_hostset_invalid_members_whitespace_element(self):
		"""Test hostset creation with whitespace-only element in members."""
		with self.assertRaises(ValueError) as context:
			self.workflow.create_hostset("hostset13", setmembers=["host1", "   "])
		self.assertIn("non-empty string", str(context.exception))
	
	def test_create_hostset_invalid_unsupported_param(self):
		"""Test hostset creation with unsupported parameter."""
		with self.assertRaises(ValueError) as context:
			self.workflow.create_hostset("hostset14", invalid_param="value")
		self.assertIn("Unsupported parameter", str(context.exception))
	
	def test_create_hostset_already_exists(self):
		"""Test hostset creation when hostset already exists."""
		self._simulate_exists("hostset_dup")
		with self.assertRaises(exceptions.HostSetAlreadyExists):
			self.workflow.create_hostset("hostset_dup")
	
	def test_create_hostset_with_none_optional_params(self):
		"""Test hostset creation with None values for optional params (should be ignored)."""
		self.workflow.create_hostset("hostset15", domain=None, comment=None, setmembers=None)
		call_args = self.session_client.rest_client.post.call_args
		payload = call_args[0][1]
		# None values should not be added to payload
		self.assertNotIn("domain", payload)
		self.assertNotIn("comment", payload)
		self.assertNotIn("members", payload)
	
	# ===================================================================
	# DELETE HOSTSET TESTS
	# ===================================================================
	
	def test_delete_hostset_success(self):
		"""Test successful hostset deletion."""
		self._simulate_exists("hostset_del", uid="uid_del")
		self.workflow.delete_hostset("hostset_del")
		self.session_client.rest_client.delete.assert_called_once_with("/hostsets/uid_del")
	
	def test_delete_hostset_not_exists(self):
		"""Test hostset deletion when hostset does not exist."""
		with self.assertRaises(exceptions.HostSetDoesNotExist):
			self.workflow.delete_hostset("nonexistent_hostset")
	
	def test_delete_hostset_invalid_name_empty(self):
		"""Test hostset deletion with empty name."""
		with self.assertRaises(ValueError):
			self.workflow.delete_hostset("")
	
	def test_delete_hostset_invalid_name_none(self):
		"""Test hostset deletion with None name."""
		with self.assertRaises(ValueError):
			self.workflow.delete_hostset(None)
	
	def test_delete_hostset_uses_uid(self):
		"""Test that delete operation uses UID in the URL, not name."""
		self._simulate_exists("hostset_uid_test", uid="unique_uid_123")
		self.workflow.delete_hostset("hostset_uid_test")
		# Verify the UID is used in the delete URL
		call_args = self.session_client.rest_client.delete.call_args
		self.assertIn("unique_uid_123", call_args[0][0])
	
	# ===================================================================
	# GET HOSTSET TESTS
	# ===================================================================
	
	def test_get_hostset_success(self):
		"""Test successful retrieval of hostset information."""
		self._simulate_exists("hostset_get", uid="uid_get", members=["host1", "host2"])
		result = self.workflow.get_hostset("hostset_get")
		self.assertEqual(result["name"], "hostset_get")
		self.assertEqual(result["uid"], "uid_get")
		self.assertEqual(result["members"], ["host1", "host2"])
	
	def test_get_hostset_with_all_properties(self):
		"""Test get hostset returns all properties."""
		self._simulate_exists("hostset_full", uid="uid_full", members=["host1"], 
							 domain="prod", comment="Production hosts")
		result = self.workflow.get_hostset("hostset_full")
		self.assertEqual(result["name"], "hostset_full")
		self.assertEqual(result["domain"], "prod")
		self.assertEqual(result["comment"], "Production hosts")
	
	def test_get_hostset_not_exists(self):
		"""Test get hostset when it doesn't exist returns empty list."""
		result = self.workflow.get_hostset("nonexistent")
		self.assertEqual(result, [])
	
	def test_get_hostset_invalid_name_empty(self):
		"""Test get hostset with empty name."""
		with self.assertRaises(ValueError):
			self.workflow.get_hostset("")
	
	def test_get_hostset_invalid_name_none(self):
		"""Test get hostset with None name."""
		with self.assertRaises(ValueError):
			self.workflow.get_hostset(None)
	
	# ===================================================================
	# HOSTSET EXISTS TESTS
	# ===================================================================
	
	def test_hostset_exists_true(self):
		"""Test hostset_exists returns True for existing hostset."""
		self._simulate_exists("hostset_exists")
		exists = self.workflow.hostset_exists("hostset_exists")
		self.assertTrue(exists)
	
	def test_hostset_exists_false(self):
		"""Test hostset_exists returns False for non-existing hostset."""
		exists = self.workflow.hostset_exists("hostset_notexists")
		self.assertFalse(exists)
	
	def test_hostset_exists_invalid_name_empty(self):
		"""Test hostset_exists with empty name."""
		with self.assertRaises(ValueError) as context:
			self.workflow.hostset_exists("")
		self.assertIn("non-empty string", str(context.exception))
	
	def test_hostset_exists_invalid_name_none(self):
		"""Test hostset_exists with None name."""
		with self.assertRaises(ValueError) as context:
			self.workflow.hostset_exists(None)
		self.assertIn("non-empty string", str(context.exception))
	
	def test_hostset_exists_invalid_name_not_string(self):
		"""Test hostset_exists with non-string name."""
		with self.assertRaises(ValueError):
			self.workflow.hostset_exists(123)
	
	def test_hostset_exists_whitespace_only(self):
		"""Test hostset_exists with whitespace-only name."""
		with self.assertRaises(ValueError):
			self.workflow.hostset_exists("   ")
	
	# ===================================================================
	# ADD HOSTS TO HOSTSET TESTS
	# ===================================================================
	
	def test_add_hosts_to_hostset_success(self):
		"""Test successfully adding hosts to hostset."""
		self._simulate_exists("hostset_add", uid="uid_add", members=["host1"])
		result = self.workflow.add_hosts_to_hostset("hostset_add", ["host2", "host3"])
		self.assertEqual(result["status"], "modified")
		# Verify PATCH was called with combined members
		call_args = self.session_client.rest_client.patch.call_args
		self.assertIn("uid_add", call_args[0][0])
		payload = call_args[1]["payload"]
		self.assertIn("host1", payload["members"])  # existing
		self.assertIn("host2", payload["members"])  # new
		self.assertIn("host3", payload["members"])  # new
	
	def test_add_hosts_to_hostset_empty_initial(self):
		"""Test adding hosts to hostset with no existing members."""
		self._simulate_exists("hostset_add2", uid="uid_add2", members=[])
		result = self.workflow.add_hosts_to_hostset("hostset_add2", ["host1"])
		self.assertEqual(result["status"], "modified")
	
	def test_add_hosts_to_hostset_single_host(self):
		"""Test adding single host to hostset."""
		self._simulate_exists("hostset_add3", uid="uid_add3", members=["host1"])
		result = self.workflow.add_hosts_to_hostset("hostset_add3", ["host2"])
		self.assertEqual(result["status"], "modified")
	
	def test_add_hosts_to_hostset_all_duplicates(self):
		"""Test adding hosts that are all already members."""
		self._simulate_exists("hostset_add4", members=["host1", "host2"])
		with self.assertRaises(exceptions.HostSetMembersAlreadyPresent):
			self.workflow.add_hosts_to_hostset("hostset_add4", ["host1", "host2"])
	
	def test_add_hosts_to_hostset_partial_duplicates(self):
		"""Test adding hosts with some duplicates (only new ones should be added)."""
		self._simulate_exists("hostset_add5", uid="uid_add5", members=["host1"])
		result = self.workflow.add_hosts_to_hostset("hostset_add5", ["host1", "host2", "host3"])
		self.assertEqual(result["status"], "modified")
		# Verify only new members are added
		call_args = self.session_client.rest_client.patch.call_args
		payload = call_args[1]["payload"]
		self.assertEqual(len(payload["members"]), 3)  # host1, host2, host3
	
	def test_add_hosts_to_hostset_not_exists(self):
		"""Test adding hosts to non-existing hostset."""
		with self.assertRaises(exceptions.HostSetDoesNotExist):
			self.workflow.add_hosts_to_hostset("nonexistent", ["host1"])
	
	def test_add_hosts_to_hostset_null_setmembers(self):
		"""Test adding null setmembers to hostset."""
		self._simulate_exists("hostset_add6")
		with self.assertRaises(ValueError) as context:
			self.workflow.add_hosts_to_hostset("hostset_add6", None)
		self.assertIn("cannot be null", str(context.exception))
	
	def test_add_hosts_to_hostset_empty_list(self):
		"""Test adding empty list of hosts."""
		self._simulate_exists("hostset_add7")
		with self.assertRaises(exceptions.HostSetMembersAlreadyPresent):
			self.workflow.add_hosts_to_hostset("hostset_add7", [])
	
	def test_add_hosts_to_hostset_invalid_name(self):
		"""Test adding hosts with invalid hostset name."""
		with self.assertRaises(ValueError):
			self.workflow.add_hosts_to_hostset("", ["host1"])
	
	def test_add_hosts_to_hostset_uses_uid(self):
		"""Test that add operation uses UID in the URL."""
		self._simulate_exists("hostset_add8", uid="unique_uid_add", members=[])
		self.workflow.add_hosts_to_hostset("hostset_add8", ["host1"])
		call_args = self.session_client.rest_client.patch.call_args
		self.assertIn("unique_uid_add", call_args[0][0])
	
	def test_add_hosts_to_hostset_preserves_order(self):
		"""Test that adding hosts preserves the order of existing members."""
		self._simulate_exists("hostset_add9", uid="uid_add9", members=["host1", "host2"])
		self.workflow.add_hosts_to_hostset("hostset_add9", ["host3"])
		call_args = self.session_client.rest_client.patch.call_args
		payload = call_args[1]["payload"]
		# Should be existing + new
		self.assertEqual(payload["members"], ["host1", "host2", "host3"])
	
	# ===================================================================
	# REMOVE HOSTS FROM HOSTSET TESTS
	# ===================================================================
	
	def test_remove_hosts_from_hostset_success(self):
		"""Test successfully removing hosts from hostset."""
		self._simulate_exists("hostset_rem", uid="uid_rem", members=["host1", "host2", "host3"])
		result = self.workflow.remove_hosts_from_hostset("hostset_rem", ["host2"])
		self.assertEqual(result["status"], "modified")
		# Verify PATCH was called with reduced members
		call_args = self.session_client.rest_client.patch.call_args
		payload = call_args[1]["payload"]
		self.assertIn("host1", payload["members"])
		self.assertIn("host3", payload["members"])
		self.assertNotIn("host2", payload["members"])
	
	def test_remove_hosts_from_hostset_multiple(self):
		"""Test removing multiple hosts from hostset."""
		self._simulate_exists("hostset_rem2", uid="uid_rem2", members=["host1", "host2", "host3", "host4"])
		result = self.workflow.remove_hosts_from_hostset("hostset_rem2", ["host2", "host3"])
		self.assertEqual(result["status"], "modified")
		call_args = self.session_client.rest_client.patch.call_args
		payload = call_args[1]["payload"]
		self.assertEqual(len(payload["members"]), 2)  # Only host1 and host4 remain
	
	def test_remove_hosts_from_hostset_all_members(self):
		"""Test removing all hosts from hostset (results in empty hostset)."""
		self._simulate_exists("hostset_rem3", uid="uid_rem3", members=["host1", "host2"])
		result = self.workflow.remove_hosts_from_hostset("hostset_rem3", ["host1", "host2"])
		self.assertEqual(result["status"], "modified")
		call_args = self.session_client.rest_client.patch.call_args
		payload = call_args[1]["payload"]
		self.assertEqual(payload["members"], [])
	
	def test_remove_hosts_from_hostset_not_member(self):
		"""Test removing hosts that are not members."""
		self._simulate_exists("hostset_rem4", members=["host1"])
		with self.assertRaises(exceptions.HostSetMembersAlreadyRemoved):
			self.workflow.remove_hosts_from_hostset("hostset_rem4", ["host2", "host3"])
	
	def test_remove_hosts_from_hostset_partial_members(self):
		"""Test removing hosts with some being members and some not."""
		self._simulate_exists("hostset_rem5", uid="uid_rem5", members=["host1", "host2"])
		result = self.workflow.remove_hosts_from_hostset("hostset_rem5", ["host1", "host3"])
		self.assertEqual(result["status"], "modified")
		# Should only remove host1 (host3 is not a member)
		call_args = self.session_client.rest_client.patch.call_args
		payload = call_args[1]["payload"]
		self.assertEqual(payload["members"], ["host2"])
	
	def test_remove_hosts_from_hostset_not_exists(self):
		"""Test removing hosts from non-existing hostset."""
		with self.assertRaises(exceptions.HostSetDoesNotExist):
			self.workflow.remove_hosts_from_hostset("nonexistent", ["host1"])
	
	def test_remove_hosts_from_hostset_null_setmembers(self):
		"""Test removing null setmembers from hostset."""
		self._simulate_exists("hostset_rem6", members=["host1"])
		with self.assertRaises(ValueError) as context:
			self.workflow.remove_hosts_from_hostset("hostset_rem6", None)
		self.assertIn("cannot be null", str(context.exception))
	
	def test_remove_hosts_from_hostset_empty_list(self):
		"""Test removing empty list of hosts."""
		self._simulate_exists("hostset_rem7", members=["host1"])
		with self.assertRaises(exceptions.HostSetMembersAlreadyRemoved):
			self.workflow.remove_hosts_from_hostset("hostset_rem7", [])
	
	def test_remove_hosts_from_hostset_invalid_name(self):
		"""Test removing hosts with invalid hostset name."""
		with self.assertRaises(ValueError):
			self.workflow.remove_hosts_from_hostset("", ["host1"])
	
	def test_remove_hosts_from_hostset_uses_uid(self):
		"""Test that remove operation uses UID in the URL."""
		self._simulate_exists("hostset_rem8", uid="unique_uid_rem", members=["host1", "host2"])
		self.workflow.remove_hosts_from_hostset("hostset_rem8", ["host1"])
		call_args = self.session_client.rest_client.patch.call_args
		self.assertIn("unique_uid_rem", call_args[0][0])
	
	def test_remove_hosts_from_hostset_preserves_order(self):
		"""Test that removing hosts preserves the order of remaining members."""
		self._simulate_exists("hostset_rem9", uid="uid_rem9", members=["host1", "host2", "host3", "host4"])
		self.workflow.remove_hosts_from_hostset("hostset_rem9", ["host2", "host4"])
		call_args = self.session_client.rest_client.patch.call_args
		payload = call_args[1]["payload"]
		# Should maintain original order of remaining members
		self.assertEqual(payload["members"], ["host1", "host3"])
	
	def test_remove_hosts_from_hostset_from_empty(self):
		"""Test removing hosts from hostset with no members."""
		self._simulate_exists("hostset_rem10", members=[])
		with self.assertRaises(exceptions.HostSetMembersAlreadyRemoved):
			self.workflow.remove_hosts_from_hostset("hostset_rem10", ["host1"])
	
	# ===================================================================
	# EDGE CASES AND BOUNDARY TESTS
	# ===================================================================
	
	def test_hostset_name_exactly_27_chars(self):
		"""Test hostset with name exactly at max length (27 characters)."""
		name_27 = "a" * 27
		self.workflow.create_hostset(name_27)
		call_args = self.session_client.rest_client.post.call_args
		self.assertEqual(call_args[0][1]["name"], name_27)
	
	def test_hostset_name_exactly_1_char(self):
		"""Test hostset with name exactly at min length (1 character)."""
		name_1 = "a"
		self.workflow.create_hostset(name_1)
		call_args = self.session_client.rest_client.post.call_args
		self.assertEqual(call_args[0][1]["name"], name_1)
	
	def test_hostset_with_special_characters_in_name(self):
		"""Test hostset with special characters in name (should be accepted if valid)."""
		name_special = "host-set_01.test"
		self.workflow.create_hostset(name_special)
		call_args = self.session_client.rest_client.post.call_args
		self.assertEqual(call_args[0][1]["name"], name_special)
	
	def test_hostset_with_large_member_list(self):
		"""Test hostset with a large number of members."""
		large_member_list = [f"host{i}" for i in range(100)]
		self.workflow.create_hostset("hostset_large", setmembers=large_member_list)
		call_args = self.session_client.rest_client.post.call_args
		self.assertEqual(len(call_args[0][1]["members"]), 100)
	
	def test_hostset_comment_empty_string_allowed(self):
		"""Test that empty string comment is allowed (not validated like domain)."""
		# Based on validator, empty comment is allowed
		self.workflow.create_hostset("hostset_empty_comment", comment="")
		call_args = self.session_client.rest_client.post.call_args
		self.assertEqual(call_args[0][1]["comment"], "")
	
	def test_multiple_operations_sequence(self):
		"""Test sequence of operations on same hostset."""
		# Create
		self.workflow.create_hostset("hostset_seq")
		# Simulate it exists now
		self._simulate_exists("hostset_seq", uid="uid_seq", members=[])
		# Add members
		self.workflow.add_hosts_to_hostset("hostset_seq", ["host1", "host2"])
		# Simulate updated members
		self._simulate_exists("hostset_seq", uid="uid_seq", members=["host1", "host2"])
		# Remove one member
		self.workflow.remove_hosts_from_hostset("hostset_seq", ["host1"])
		# Simulate updated members
		self._simulate_exists("hostset_seq", uid="uid_seq", members=["host2"])
		# Get info
		result = self.workflow.get_hostset("hostset_seq")
		self.assertEqual(result["members"], ["host2"])
	
	def test_rest_client_exception_handling(self):
		"""Test that REST client exceptions are properly propagated."""
		self.session_client.rest_client.get.side_effect = Exception("Connection error")
		with self.assertRaises(Exception) as context:
			self.workflow.hostset_exists("any_hostset")
		self.assertIn("Connection error", str(context.exception))
	
	def test_create_hostset_with_mixed_case_name(self):
		"""Test hostset creation with mixed case name."""
		self.workflow.create_hostset("HostSet-MixedCase")
		call_args = self.session_client.rest_client.post.call_args
		self.assertEqual(call_args[0][1]["name"], "HostSet-MixedCase")
	
	def test_unicode_characters_in_comment(self):
		"""Test hostset with unicode characters in comment."""
		unicode_comment = "Test hostset with émojis 🚀 and spëcial chars"
		self.workflow.create_hostset("hostset_unicode", comment=unicode_comment)
		call_args = self.session_client.rest_client.post.call_args
		self.assertEqual(call_args[0][1]["comment"], unicode_comment)


if __name__ == "__main__":
	unittest.main()



#    (c) Copyright 2026 Hewlett Packard Enterprise Development LP
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from typing import Any, List, Dict

from hpe_storage_flowkit_py.v3.src.core.session import SessionManager
from hpe_storage_flowkit_py.v3.src.validators.vlun_validator import validate_vlun_params
from hpe_storage_flowkit_py.v3.src.validators.vlun_validator import validate_vlun_optional_params
from hpe_storage_flowkit_py.v3.src.core.logger import Logger
logger = Logger()
class VLunWorkflow:

	def __init__(self, session_mgr: SessionManager):
		self.session_mgr = session_mgr

	def _execute_create_vlun(self, volname, hostname, **kwargs):
		"""
		Execute vlun creation with payload for HPE 3PAR SSMC API.
		
		Required parameters:
		- volname: Name of the volume to be exported
		- hostname: Name of the host to export to.
		"""
		try:
			# Validate required parameters
			validate_vlun_params(volname, hostname)
			
			# Build the base payload with required fields
			payload = {
				"volumeName": volname,
				"hostName": hostname
			}
			# Process optional parameters directly from kwargs
			opt_params = {}
			if kwargs:
				opt_params = {k: v for k, v in kwargs.items() if v is not None}
			if opt_params:
				validate_vlun_optional_params(opt_params)
			
			payload.update(opt_params)
			logger.info(f"VLun creation payload: {payload}")
			
			response = self.session_mgr.rest_client.post(f"/vluns", payload)
			return response
		except Exception as e:
			# Log the error
			logger.error(f"VLun creation failed: {e}")
			raise


	def create_vlun(self, volname, hostname, **kwargs):
		try:
			response = self._execute_create_vlun(volname, hostname, **kwargs)
			return response
		except Exception as e:
			raise

	
	def delete_vlun(self, volname, lun, hostname=None, port=None):
		uids = self._get_vlun_uids(volname, lun, hostname, port)
		logger.info(f"Retrieved VLUN UIDs: {uids}")
		responses = []
		if isinstance(uids, list) and uids:
			for uid in uids:
				try:
					response = self.session_mgr.rest_client.delete(f"/vluns/{uid}")
					logger.info(f"Deleted VLUN with UID: {uid}")
					responses.append(response)

				except Exception as e:
					logger.error(f"VLUN deletion failed for UID {uid}: {e}")
					raise
			return responses
		else:
			logger.error(f"VLUN UIDs not found for volume '{volname}' and LUN '{lun}'")
			return None
		
    	
	def _get_vlun_uids(self, volname, lun, hostname, port):
		if volname is not None and lun is not None:
			try:
				query = f"volumeName={volname}&lun={str(lun)}"
				if hostname is not None:
					query += f"&hostName={hostname}"
				if port is not None:
					query += f"&portPos={port}"
				
				headers = {'experimentalfilter': 'true'}
				uri = f"/vluns?{query}"
				response = self.session_mgr.rest_client.get(uri, headers=headers)

				members = self.normalize_members(response)
				if members:
					return [item["uid"] for item in response if "uid" in item]

			except Exception as e:
				logger.error(f"Fetching VLun UIDs failed: {e}")
				raise
		else:
			return None


	def get_vlun(self, queryKey, queryValue):
		if queryKey is not None and queryValue is not None:
			try:
				headers = {'experimentalfilter': 'true'}
				uri = f"/vluns?{queryKey}={queryValue}"
				response = self.session_mgr.rest_client.get(uri, headers=headers)
				members = self.normalize_members(response)
				if not members:
					raise ValueError("Invalid vlun response format")

				return members
			except Exception as e:
				logger.error(f"Fetching VLun failed: {e}")
				raise
		return None
	

	def get_vluns(self):
		try:
			response = self.session_mgr.rest_client.get(f"/vluns")
			members = self.normalize_members(response)
			if not members:
				raise ValueError("Invalid vluns response format")
			return members
		except Exception as e:
				logger.error(f"Fetching VLuns failed: {e}")
				raise


	@staticmethod
	def normalize_members(response: Any) -> List[Dict]:
		"""
		Normalize API responses that may be in one of several shapes:
    		- dict with key "members" as dict: return list(members.values())
    		- dict with key "members" as list: return that list
    		- plain list: return as-is
    		- otherwise: return empty list
    	"""
		if isinstance(response, dict):
			members = response.get("members")
			if isinstance(members, dict):
				return list(members.values())
			if isinstance(members, list):
				return members
			return []
		if isinstance(response, list):
			return response
		return []

	




#    (c) Copyright 2026 Hewlett Packard Enterprise Development LP
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import unittest
from unittest.mock import Mock
import sys
import os

# Add the src directory to the path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from hpe_storage_flowkit_py.v3.src.workflows.port import PortWorkflow
from hpe_storage_flowkit_py.v3.src.core.exceptions import HPEStorageException


class MockSessionManager:
    """Mock session manager for testing"""
    def __init__(self):
        self.rest_client = MockRestClient()


class MockRestClient:
    """Mock REST client for testing"""
    def __init__(self):
        self.get_response = None
    
    def get(self, endpoint):
        """Mock GET request"""
        if self.get_response is not None:
            return self.get_response
        
        # Default response - ports as dict of dicts
        return {
            'members': {
                'port-1': {
                    'portPos': {'node': 0, 'slot': 0, 'cardPort': 1},
                    'type': 'HOST',
                    'protocol': 'FC',
                    'mode': 'TARGET',
                    'linkState': 'READY'
                },
                'port-2': {
                    'portPos': {'node': 1, 'slot': 1, 'cardPort': 1},
                    'type': 'HOST',
                    'protocol': 'FC',
                    'mode': 'TARGET',
                    'linkState': 'READY'
                }
            }
        }


class TestPortWorkflow(unittest.TestCase):
    """Unit tests for PortWorkflow class"""
    
    def setUp(self):
        """Set up test fixtures"""
        self.session_mgr = MockSessionManager()
        self.workflow = PortWorkflow(self.session_mgr)
    
    def test_get_ports_success_dict_members(self):
        """Test getting ports when response has members as dict"""
        result = self.workflow.get_ports()
        
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0]['type'], 'HOST')
        self.assertEqual(result[0]['protocol'], 'FC')
    
    def test_get_ports_success_list_members(self):
        """Test getting ports when response has members as list"""
        self.session_mgr.rest_client.get_response = {
            'members': [
                {'portPos': {'node': 0}, 'type': 'HOST'},
                {'portPos': {'node': 1}, 'type': 'HOST'}
            ]
        }
        
        result = self.workflow.get_ports()
        
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)
    
    def test_get_ports_direct_list_response(self):
        """Test getting ports when response is a direct list"""
        self.session_mgr.rest_client.get_response = [
            {'portPos': {'node': 0}, 'type': 'HOST'},
            {'portPos': {'node': 1}, 'type': 'DISK'}
        ]
        
        result = self.workflow.get_ports()
        
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0]['type'], 'HOST')
        self.assertEqual(result[1]['type'], 'DISK')
    
    def test_get_ports_empty_members_dict(self):
        """Test getting ports when members dict is empty"""
        self.session_mgr.rest_client.get_response = {'members': {}}
        
        with self.assertRaises(ValueError) as context:
            self.workflow.get_ports()
        
        self.assertIn('Invalid ports response', str(context.exception))
    
    def test_get_ports_empty_members_list(self):
        """Test getting ports when members list is empty"""
        self.session_mgr.rest_client.get_response = {'members': []}
        
        with self.assertRaises(ValueError) as context:
            self.workflow.get_ports()
        
        self.assertIn('Invalid ports response', str(context.exception))
    
    def test_get_ports_no_members_key(self):
        """Test getting ports when response has no members key"""
        self.session_mgr.rest_client.get_response = {'data': 'some_data'}
        
        with self.assertRaises(ValueError) as context:
            self.workflow.get_ports()
        
        self.assertIn('Invalid ports response', str(context.exception))
    
    def test_get_ports_exception_handling(self):
        """Test exception handling in get_ports"""
        def raise_exception(endpoint):
            raise Exception("Network error")
        
        self.session_mgr.rest_client.get = raise_exception
        
        with self.assertRaises(Exception) as context:
            self.workflow.get_ports()
        
        self.assertIn('Network error', str(context.exception))
    
    def test_normalize_members_dict_of_dicts(self):
        """Test normalize_members with dict containing dict members"""
        response = {
            'members': {
                'port-1': {'name': 'port1'},
                'port-2': {'name': 'port2'}
            }
        }
        
        result = PortWorkflow.normalize_members(response)
        
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)
    
    def test_normalize_members_dict_of_list(self):
        """Test normalize_members with dict containing list members"""
        response = {
            'members': [
                {'name': 'port1'},
                {'name': 'port2'}
            ]
        }
        
        result = PortWorkflow.normalize_members(response)
        
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0]['name'], 'port1')
    
    def test_normalize_members_direct_list(self):
        """Test normalize_members with direct list"""
        response = [
            {'name': 'port1'},
            {'name': 'port2'}
        ]
        
        result = PortWorkflow.normalize_members(response)
        
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)
    
    def test_normalize_members_empty_dict(self):
        """Test normalize_members with empty dict"""
        result = PortWorkflow.normalize_members({})
        
        self.assertEqual(result, [])
    
    def test_normalize_members_dict_no_members(self):
        """Test normalize_members with dict but no members key"""
        result = PortWorkflow.normalize_members({'data': 'something'})
        
        self.assertEqual(result, [])
    
    def test_normalize_members_none(self):
        """Test normalize_members with None"""
        result = PortWorkflow.normalize_members(None)
        
        self.assertEqual(result, [])
    
    def test_normalize_members_string(self):
        """Test normalize_members with string"""
        result = PortWorkflow.normalize_members("not a dict or list")
        
        self.assertEqual(result, [])
    
    def test_normalize_members_empty_list(self):
        """Test normalize_members with empty list"""
        result = PortWorkflow.normalize_members([])
        
        self.assertEqual(result, [])
    
    def test_get_ports_multiple_port_types(self):
        """Test getting ports with multiple port types"""
        self.session_mgr.rest_client.get_response = {
            'members': {
                'port-1': {'type': 'HOST', 'protocol': 'FC'},
                'port-2': {'type': 'DISK', 'protocol': 'SAS'},
                'port-3': {'type': 'HOST', 'protocol': 'iSCSI'},
                'port-4': {'type': 'RCFC', 'protocol': 'FC'}
            }
        }
        
        result = self.workflow.get_ports()
        
        self.assertEqual(len(result), 4)
        types = [port['type'] for port in result]
        self.assertIn('HOST', types)
        self.assertIn('DISK', types)
        self.assertIn('RCFC', types)


if __name__ == '__main__':
    unittest.main()




#    (c) Copyright 2026 Hewlett Packard Enterprise Development LP
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Validator functions for Host Set operations.

Rules:
    - `name` always required.
    - Supported optional kwargs: domain, comment, setmembers.
    - Unknown kwargs raise ValueError.
"""

from hpe_storage_flowkit_py.v3.src.core.logger import Logger

logger = Logger()

NAME_MIN = 1
NAME_MAX = 27
OPTIONAL_PARAMS = {"domain", "comment", "setmembers"}


def validate_hostset_params(name: str, **kwargs) -> bool:
    """Validate host set parameters.

    Parameters:
      name: host set name (required).
      kwargs: optional params (domain, comment, setmembers).
    
    Raises:
      ValueError: When any parameter is invalid.
    
    Returns:
      bool: True if all validations pass.
    """
    logger.info("Started validating host set parameters")
    # Validate required: name
    if name is None:
        logger.error("Host set name parameter is null")
        raise ValueError('Host set name cannot be null.')
    if not isinstance(name, str) or not name.strip():
        logger.error(f"Host set name is not a valid string: {name}")
        raise ValueError('Host set name must be a non-empty string.')
    if not (NAME_MIN <= len(name) <= NAME_MAX):
        logger.error(f"Host set name length ({len(name)}) is outside valid range {NAME_MIN}-{NAME_MAX}")
        raise ValueError(f'Host set name must be {NAME_MIN}-{NAME_MAX} characters in length.')

    # Iterate kwargs for validation
    for key, value in kwargs.items():
        if key not in OPTIONAL_PARAMS:
            logger.error(f"Unsupported parameter '{key}' for host set operation")
            raise ValueError(f"Unsupported parameter for host set operation: {key}")
        
        if key == "domain":
            if value is not None:
                if not isinstance(value, str):
                    logger.error(f"Domain parameter must be a string, got {type(value)}")
                    raise ValueError('Domain must be a string when provided.')
                if not value.strip():
                    logger.error("Domain parameter cannot be empty")
                    raise ValueError('Domain cannot be empty when provided.')
        
        elif key == "comment":
            if value is not None:
                if not isinstance(value, str):
                    logger.error(f"Comment parameter must be a string, got {type(value)}")
                    raise ValueError('Comment must be a string when provided.')
        
        elif key == "setmembers":
            if value is not None:
                if not isinstance(value, list):
                    logger.error(f"SetMembers parameter must be a list, got {type(value)}")
                    raise ValueError(f'setmembers must be a list of host names. Current provided type is {type(value)}.')
                for h in value:
                    if not isinstance(h, str) or not h.strip():
                        logger.error(f"Invalid member in setmembers list: {h}")
                        raise ValueError('Each set member must be a non-empty string.')
    
    logger.info("Completed validating host set parameters")
    return True



#    (c) Copyright 2026 Hewlett Packard Enterprise Development LP
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import unittest
from unittest.mock import MagicMock, Mock, patch
import sys
import os

# Ensure src is on sys.path for direct imports
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from hpe_storage_flowkit_py.v3.src.workflows.user import UserWorkflow
from hpe_storage_flowkit_py.v3.src.workflows.task import TaskManager
from hpe_storage_flowkit_py.v3.src.core import exceptions
from hpe_storage_flowkit_py.v3.src.core.rest_client import RESTClient
from hpe_storage_flowkit_py.v3.src.validators.user_validator import validate_password_change_params, validate_user_operation_params


class MockTaskManager:
	"""Mock TaskManager for testing user workflows."""
	def __init__(self, session_mgr):
		self.session_mgr = session_mgr
	
	def wait_for_task_to_end(self, task_uri):
		return {"status": "STATE_FINISHED", "state": {"overall": "STATE_NORMAL"}}


class TestUserWorkflow(unittest.TestCase):
	"""Unit tests for UserWorkflow class using a simple Mock REST client.

	Tests cover all methods with positive and negative scenarios including edge cases.
	"""

	def setUp(self):
		"""Set up test fixtures before each test method."""
		self.session_client = Mock()
		self.session_client.rest_client = Mock(spec=RESTClient)
		# Default canned responses (no state tracking)
		self.session_client.rest_client.get.return_value = {"members": {}}
		self.session_client.rest_client.post.return_value = {"status": "created"}
		self.session_client.rest_client.delete.return_value = {"status": "deleted"}
		self.session_client.rest_client.patch.return_value = {"status": "modified"}
		self.task_mgr = MockTaskManager(self.session_client)
		self.workflow = UserWorkflow(self.session_client, self.task_mgr)

	def _simulate_user_exists(self, name, uid="uid_x", domain_privileges=None):
		"""Helper to simulate an existing user with given properties."""
		domain_privileges = domain_privileges or [{"name": "default", "privilege": "browse"}]
		user_data = {
			"uid": uid,
			"name": name,
			"domainPrivileges": domain_privileges
		}
		
		self.session_client.rest_client.get.return_value = {
			"members": {
				uid: user_data
			}
		}


	
	def test_create_user_success_minimal(self):
		"""Test successful user creation with minimal required parameters."""
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		self.workflow.create_user("testuser", "password123", domain_privileges)
		self.session_client.rest_client.post.assert_called_once()
		call_args = self.session_client.rest_client.post.call_args
		self.assertEqual(call_args[0][0], "/users")
		payload = call_args[0][1]
		self.assertEqual(payload["name"], "testuser")
		self.assertEqual(payload["password"], ['p', 'a', 's', 's', 'w', 'o', 'r', 'd', '1', '2', '3'])
		# Check domain privileges structure
		self.assertEqual(len(payload["domainPrivileges"]), 1)
		self.assertEqual(payload["domainPrivileges"][0]["name"], "default")
		self.assertEqual(payload["domainPrivileges"][0]["priv"], "browse")
	
	def test_create_user_success_with_all_params(self):
		"""Test user creation with all parameters."""
		domain_privileges = [
			{"name": "production", "privilege": "edit"},
			{"name": "development", "privilege": "browse"}
		]
		self.workflow.create_user(
			"admin_user",
			"SecurePass123!",
			domain_privileges
		)
		call_args = self.session_client.rest_client.post.call_args
		payload = call_args[0][1]
		self.assertEqual(payload["name"], "admin_user")
		self.assertEqual(payload["password"], ['S', 'e', 'c', 'u', 'r', 'e', 'P', 'a', 's', 's', '1', '2', '3', '!'])
		# Check domain privileges structure
		self.assertEqual(len(payload["domainPrivileges"]), 2)
		self.assertEqual(payload["domainPrivileges"][0]["name"], "production")
		self.assertEqual(payload["domainPrivileges"][0]["priv"], "edit")
		self.assertEqual(payload["domainPrivileges"][1]["name"], "development")
		self.assertEqual(payload["domainPrivileges"][1]["priv"], "browse")
	
	def test_create_user_async_task_202(self):
		"""Test user creation with async task (202 response)."""
		# Mock task manager for this test
		mock_task_manager = MagicMock()
		mock_task_manager.wait_for_task_to_end.return_value = {"status": "completed"}
		self.workflow.task_manager = mock_task_manager
		
		# Mock 202 response with resourceUri
		self.session_client.rest_client.post.return_value = {
			"message": "Started task to execute create User",
			"resourceUri": "/api/v3/tasks/abc123"
		}
		
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		self.workflow.create_user("async_user", "password123", domain_privileges)
		
		mock_task_manager.wait_for_task_to_end.assert_called_once_with(
			"/api/v3/tasks/abc123"
		)
	
	def test_create_user_invalid_name_empty(self):
		"""Test user creation with empty name."""
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("", "password123", domain_privileges)
		self.assertIn("cannot be empty", str(context.exception))
	
	def test_create_user_invalid_name_none(self):
		"""Test user creation with None name."""
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user(None, "password123", domain_privileges)
		self.assertIn("User name is required", str(context.exception))
	
	def test_create_user_invalid_name_non_string(self):
		"""Test user creation with non-string name."""
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user(123, "password123", domain_privileges)
		self.assertIn("must be a string", str(context.exception))
	
	def test_create_user_invalid_name_with_special_chars(self):
		"""Test user creation with invalid characters in name."""
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("user/invalid", "password123", domain_privileges)
		self.assertIn("invalid characters", str(context.exception))
	
	def test_create_user_invalid_password_empty(self):
		"""Test user creation with empty password."""
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", "", domain_privileges)
		self.assertIn("cannot be empty", str(context.exception))
	
	def test_create_user_invalid_password_too_short(self):
		"""Test user creation with password too short."""
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", "short", domain_privileges)
		self.assertIn("at least 8 characters", str(context.exception))
	
	def test_create_user_invalid_password_non_string_list(self):
		"""Test user creation with invalid password type."""
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", 12345, domain_privileges)
		self.assertIn("must be a string or list", str(context.exception))
	
	def test_create_user_invalid_domain_privileges_empty(self):
		"""Test user creation with empty domain privileges."""
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", "password123", [])
		self.assertIn("At least one domain privilege must be specified", str(context.exception))
	
	def test_create_user_invalid_domain_privileges_not_list(self):
		"""Test user creation with domain privileges not being a list."""
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", "password123", "invalid")
		self.assertIn("must be a list", str(context.exception))
	
	def test_create_user_invalid_domain_privilege_missing_domain(self):
		"""Test user creation with domain privilege missing domain."""
		domain_privileges = [{"privilege": "browse"}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", "password123", domain_privileges)
		self.assertIn("is missing 'name' field", str(context.exception))
	
	def test_create_user_invalid_domain_privilege_missing_privilege(self):
		"""Test user creation with domain privilege missing privilege."""
		domain_privileges = [{"name": "default"}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", "password123", domain_privileges)
		self.assertIn("is missing 'privilege' field", str(context.exception))
	
	def test_create_user_invalid_privilege_value(self):
		"""Test user creation with invalid privilege value."""
		domain_privileges = [{"name": "default", "privilege": "invalid_privilege"}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", "password123", domain_privileges)
		self.assertIn("Invalid privilege", str(context.exception))
	
	def test_create_user_already_exists(self):
		"""Test user creation when user already exists."""
		self._simulate_user_exists("duplicate_user")
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		with self.assertRaises(exceptions.UserAlreadyExists):
			self.workflow.create_user("duplicate_user", "password123", domain_privileges)
	
	def test_create_user_invalid_unsupported_param(self):
		"""Test user creation with unsupported parameter - should be ignored."""
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		# Unsupported parameters are ignored, not rejected
		self.workflow.create_user("testuser", "password123", domain_privileges, invalid_param="value")
		self.session_client.rest_client.post.assert_called_once()
	

	
	def test_get_all_users_success_empty(self):
		"""Test successful retrieval of all users when none exist."""
		result = self.workflow.get_all_users()
		self.assertEqual(result, {"members": {}})
		self.session_client.rest_client.get.assert_called_once_with("/users")
	
	def test_get_all_users_success_multiple(self):
		"""Test successful retrieval of multiple users."""
		self.session_client.rest_client.get.return_value = {
			"members": {
				"uid1": {"uid": "uid1", "name": "user1", "domainPrivileges": [{"name": "default", "privilege": "browse"}]},
				"uid2": {"uid": "uid2", "name": "user2", "domainPrivileges": [{"name": "prod", "privilege": "edit"}]}
			}
		}
		result = self.workflow.get_all_users()
		# get_all_users returns the raw API response
		self.assertIn("members", result)
		self.assertEqual(len(result["members"]), 2)
		users = list(result["members"].values())
		self.assertEqual(users[0]["name"], "user1")
		self.assertEqual(users[1]["name"], "user2")
	
	def test_get_all_users_success_single(self):
		"""Test successful retrieval of single user."""
		self._simulate_user_exists("single_user")
		result = self.workflow.get_all_users()
		# get_all_users returns the raw API response
		self.assertIn("members", result)
		self.assertEqual(len(result["members"]), 1)
		users = list(result["members"].values())
		self.assertEqual(users[0]["name"], "single_user")
	

	
	def test_get_user_by_name_success(self):
		"""Test successful retrieval of user by name."""
		self._simulate_user_exists("test_user", uid="uid_test")
		result = self.workflow.get_user_by_name("test_user")
		self.assertEqual(result["name"], "test_user")
		self.assertEqual(result["uid"], "uid_test")
	
	def test_get_user_by_name_not_exists(self):
		"""Test get user by name when user doesn't exist."""
		with self.assertRaises(exceptions.UserDoesNotExist):
			self.workflow.get_user_by_name("nonexistent_user")
	
	def test_get_user_by_name_invalid_name_empty(self):
		"""Test get user by name with empty name."""
		with self.assertRaises(exceptions.InvalidInput):
			self.workflow.get_user_by_name("")
	
	def test_get_user_by_name_invalid_name_none(self):
		"""Test get user by name with None name."""
		with self.assertRaises(exceptions.InvalidInput):
			self.workflow.get_user_by_name(None)
	
	def test_get_user_by_name_invalid_name_non_string(self):
		"""Test get user by name with non-string name."""
		with self.assertRaises(exceptions.InvalidInput):
			self.workflow.get_user_by_name(123)
	

	
	def test_modify_user_success_password_only(self):
		"""Test successful user modification with password change only."""
		self._simulate_user_exists("modify_user", uid="uid_modify")
		self.workflow.modify_user_by_name(
			"modify_user",
			current_password="oldpass123",
			new_password="newpass456"
		)
		self.session_client.rest_client.patch.assert_called_once()
		call_args = self.session_client.rest_client.patch.call_args
		self.assertEqual(call_args[0][0], "/users/uid_modify")
		payload = call_args[0][1]
		self.assertEqual(payload["currentPassword"], ['o', 'l', 'd', 'p', 'a', 's', 's', '1', '2', '3'])
		self.assertEqual(payload["newPassword"], ['n', 'e', 'w', 'p', 'a', 's', 's', '4', '5', '6'])
	
	def test_modify_user_success_domain_privileges_only(self):
		"""Test successful user modification with domain privileges change only."""
		self._simulate_user_exists("modify_user", uid="uid_modify")
		new_privileges = [{"name": "production", "privilege": "edit"}]
		self.workflow.modify_user_by_name("modify_user", domain_privileges=new_privileges)
		call_args = self.session_client.rest_client.patch.call_args
		payload = call_args[0][1]
		# Check transformed domain privileges
		self.assertEqual(len(payload["domainPrivileges"]), 1)
		self.assertEqual(payload["domainPrivileges"][0]["name"], "production")
		self.assertEqual(payload["domainPrivileges"][0]["priv"], "edit")
		self.assertNotIn("currentPassword", payload)
		self.assertNotIn("newPassword", payload)
	
	def test_modify_user_success_all_params(self):
		"""Test successful user modification with all parameters."""
		self._simulate_user_exists("modify_user", uid="uid_modify")
		new_privileges = [{"name": "test", "privilege": "super"}]
		self.workflow.modify_user_by_name(
			"modify_user",
			current_password="current123",
			new_password="newpass456",
			domain_privileges=new_privileges
		)
		call_args = self.session_client.rest_client.patch.call_args
		payload = call_args[0][1]
		self.assertEqual(payload["currentPassword"], ['c', 'u', 'r', 'r', 'e', 'n', 't', '1', '2', '3'])
		self.assertEqual(payload["newPassword"], ['n', 'e', 'w', 'p', 'a', 's', 's', '4', '5', '6'])
		# Check transformed domain privileges
		self.assertEqual(len(payload["domainPrivileges"]), 1)
		self.assertEqual(payload["domainPrivileges"][0]["name"], "test")
		self.assertEqual(payload["domainPrivileges"][0]["priv"], "super")
	
	def test_modify_user_not_exists(self):
		"""Test modify user when user doesn't exist."""
		with self.assertRaises(exceptions.UserDoesNotExist):
			self.workflow.modify_user_by_name("nonexistent", 
				current_password="current123", 
				new_password="newpass123")
	
	def test_modify_user_no_changes(self):
		"""Test modify user with no actual changes provided."""
		self._simulate_user_exists("modify_user", uid="uid_modify")
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.modify_user_by_name("modify_user")
		self.assertIn("At least one parameter", str(context.exception))
	
	def test_modify_user_invalid_password_combination(self):
		"""Test modify user with invalid password combination (only current password provided)."""
		self._simulate_user_exists("modify_user", uid="uid_modify")
		# Only providing current_password is allowed - API validation would catch issues
		self.workflow.modify_user_by_name("modify_user", current_password="current123")
		self.session_client.rest_client.patch.assert_called_once()
	
	def test_modify_user_invalid_new_password_only(self):
		"""Test modify user with only new password provided."""
		self._simulate_user_exists("modify_user", uid="uid_modify")
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.modify_user_by_name("modify_user", new_password="newpass123")
		self.assertIn("currentPassword is required when changing password", str(context.exception))
	
	def test_modify_user_invalid_name_empty(self):
		"""Test modify user with empty name."""
		with self.assertRaises(exceptions.InvalidInput):
			self.workflow.modify_user_by_name("", new_password="newpass123")
	
	def test_modify_user_invalid_name_none(self):
		"""Test modify user with None name."""
		with self.assertRaises(exceptions.InvalidInput):
			self.workflow.modify_user_by_name(None, new_password="newpass123")
	
	def test_modify_user_invalid_unsupported_param(self):
		"""Test modify user with unsupported parameter - should be ignored."""
		self._simulate_user_exists("modify_user", uid="uid_modify")
		# Unsupported parameters are ignored, not rejected, but still need valid params
		self.workflow.modify_user_by_name("modify_user", 
			current_password="current123", 
			new_password="newpass123", 
			invalid_param="value")
		self.session_client.rest_client.patch.assert_called_once()
	

	
	def test_delete_user_success(self):
		"""Test successful user deletion."""
		self._simulate_user_exists("delete_user", uid="uid_delete")
		self.workflow.delete_user_by_name("delete_user")
		self.session_client.rest_client.delete.assert_called_once_with("/users/uid_delete")
	
	def test_delete_user_not_exists(self):
		"""Test user deletion when user does not exist."""
		with self.assertRaises(exceptions.UserDoesNotExist):
			self.workflow.delete_user_by_name("nonexistent_user")
	
	def test_delete_user_invalid_name_empty(self):
		"""Test user deletion with empty name."""
		with self.assertRaises(exceptions.InvalidInput):
			self.workflow.delete_user_by_name("")
	
	def test_delete_user_invalid_name_none(self):
		"""Test user deletion with None name."""
		with self.assertRaises(exceptions.InvalidInput):
			self.workflow.delete_user_by_name(None)
	
	def test_delete_user_invalid_name_non_string(self):
		"""Test user deletion with non-string name."""
		with self.assertRaises(exceptions.InvalidInput):
			self.workflow.delete_user_by_name(123)
	
	def test_delete_user_uses_uid(self):
		"""Test that delete operation uses UID in the URL, not name."""
		self._simulate_user_exists("delete_user_uid", uid="unique_uid_delete")
		self.workflow.delete_user_by_name("delete_user_uid")
		# Verify the UID is used in the delete URL
		call_args = self.session_client.rest_client.delete.call_args
		self.assertIn("unique_uid_delete", call_args[0][0])
	

	
	def test_fetch_user_uid_by_name_success(self):
		"""Test successful UID retrieval by name."""
		self._simulate_user_exists("test_uid_user", uid="special_uid_123")
		uid = self.workflow._fetch_user_uid_by_name("test_uid_user")
		self.assertEqual(uid, "special_uid_123")
	
	def test_fetch_user_uid_by_name_not_exists(self):
		"""Test UID retrieval when user doesn't exist."""
		with self.assertRaises(exceptions.UserDoesNotExist):
			self.workflow._fetch_user_uid_by_name("nonexistent_user")
	
	def test_build_create_user_payload(self):
		"""Test building create user payload."""
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		payload = self.workflow._build_create_user_payload("testuser", "password123", domain_privileges)
		self.assertEqual(payload["name"], "testuser")
		self.assertEqual(payload["password"], ['p', 'a', 's', 's', 'w', 'o', 'r', 'd', '1', '2', '3'])
		# The API transforms domain privileges structure - check first entry
		self.assertEqual(len(payload["domainPrivileges"]), 1)
		self.assertEqual(payload["domainPrivileges"][0]["name"], "default")
		self.assertEqual(payload["domainPrivileges"][0]["priv"], "browse")
	
	def test_build_modify_user_payload_password_only(self):
		"""Test building modify user payload with password change."""
		payload = self.workflow._build_modify_user_payload(
			current_password="old123",
			new_password="newpass456"
		)
		self.assertEqual(payload["currentPassword"], ['o', 'l', 'd', '1', '2', '3'])
		self.assertEqual(payload["newPassword"], ['n', 'e', 'w', 'p', 'a', 's', 's', '4', '5', '6'])
		self.assertNotIn("domainPrivileges", payload)
	
	def test_build_modify_user_payload_privileges_only(self):
		"""Test building modify user payload with domain privileges only."""
		domain_privileges = [{"name": "test", "privilege": "edit"}]
		payload = self.workflow._build_modify_user_payload(domain_privileges=domain_privileges)
		# Check transformed domain privileges structure
		self.assertEqual(len(payload["domainPrivileges"]), 1)
		self.assertEqual(payload["domainPrivileges"][0]["name"], "test")
		self.assertEqual(payload["domainPrivileges"][0]["priv"], "edit")
		self.assertNotIn("currentPassword", payload)
		self.assertNotIn("newPassword", payload)
	
	def test_build_modify_user_payload_all_params(self):
		"""Test building modify user payload with all parameters."""
		domain_privileges = [{"name": "prod", "privilege": "super"}]
		payload = self.workflow._build_modify_user_payload(
			current_password="current123",
			new_password="newpass456",
			domain_privileges=domain_privileges
		)
		self.assertEqual(payload["currentPassword"], ['c', 'u', 'r', 'r', 'e', 'n', 't', '1', '2', '3'])
		self.assertEqual(payload["newPassword"], ['n', 'e', 'w', 'p', 'a', 's', 's', '4', '5', '6'])
		# Check transformed domain privileges structure
		self.assertEqual(len(payload["domainPrivileges"]), 1)
		self.assertEqual(payload["domainPrivileges"][0]["name"], "prod")
		self.assertEqual(payload["domainPrivileges"][0]["priv"], "super")

	# ===================================================================
	# PASSWORD LIST FORMAT TESTS
	# ===================================================================

	def test_create_user_password_as_list(self):
		"""Test user creation with password provided as list of characters."""
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		password_list = ['p', 'a', 's', 's', 'w', 'o', 'r', 'd', '1', '2', '3']
		self.workflow.create_user("testuser", password_list, domain_privileges)
		call_args = self.session_client.rest_client.post.call_args
		payload = call_args[0][1]
		self.assertEqual(payload["password"], password_list)

	def test_modify_user_password_as_list(self):
		"""Test user modification with passwords provided as lists."""
		self._simulate_user_exists("modify_user", uid="uid_modify")
		current_pwd = ['o', 'l', 'd', 'p', 'a', 's', 's', '1', '2', '3']
		new_pwd = ['n', 'e', 'w', 'p', 'a', 's', 's', '4', '5', '6']
		self.workflow.modify_user_by_name("modify_user", current_password=current_pwd, new_password=new_pwd)
		call_args = self.session_client.rest_client.patch.call_args
		payload = call_args[0][1]
		self.assertEqual(payload["currentPassword"], current_pwd)
		self.assertEqual(payload["newPassword"], new_pwd)

	# ===================================================================
	# ASYNC TASK HANDLING TESTS
	# ===================================================================

	def test_create_user_with_task_uri(self):
		"""Test user creation that returns a task URI."""
		self.session_client.rest_client.post.return_value = {"taskUri": "/api/v3/tasks/task_123"}
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		
		result = self.workflow.create_user("testuser", "password123", domain_privileges)
		
		self.assertEqual(result["status"], "STATE_FINISHED")

	def test_modify_user_with_resource_uri(self):
		"""Test user modification that returns a resource URI."""
		self._simulate_user_exists("modify_user", uid="uid_modify")
		self.session_client.rest_client.patch.return_value = {"resourceUri": "/api/v3/users/uid_modify"}
		
		result = self.workflow.modify_user_by_name("modify_user", 
			current_password="oldpass123", new_password="newpass456")
		
		self.assertEqual(result["status"], "STATE_FINISHED")

	def test_delete_user_with_task_uri(self):
		"""Test user deletion that returns a task URI."""
		self._simulate_user_exists("delete_user", uid="uid_delete")
		self.session_client.rest_client.delete.return_value = {"taskUri": "/api/v3/tasks/task_456"}
		
		result = self.workflow.delete_user_by_name("delete_user")
		
		self.assertEqual(result["status"], "STATE_FINISHED")

	# ===================================================================
	# PARSE USERS RESPONSE TESTS
	# ===================================================================

	def test_parse_users_response_dict_with_members_dict(self):
		"""Test parsing response with members as dict."""
		response = {
			"members": {
				"uid1": {"name": "user1", "uid": "uid1"},
				"uid2": {"name": "user2", "uid": "uid2"}
			}
		}
		result = self.workflow._parse_users_response(response)
		self.assertEqual(len(result), 2)
		self.assertEqual(result[0]["name"], "user1")

	def test_parse_users_response_plain_list(self):
		"""Test parsing response that is a plain list already."""
		response = [
			{"name": "user1", "uid": "uid1"},
			{"name": "user2", "uid": "uid2"}
		]
		result = self.workflow._parse_users_response(response)
		self.assertEqual(len(result), 2)
		self.assertEqual(result[0]["name"], "user1")

	def test_parse_users_response_plain_list(self):
		"""Test parsing response that is a plain list."""
		response = [
			{"name": "user1", "uid": "uid1"},
			{"name": "user2", "uid": "uid2"}
		]
		result = self.workflow._parse_users_response(response)
		self.assertEqual(len(result), 2)
		self.assertEqual(result[0]["name"], "user1")

	def test_parse_users_response_single_user_dict(self):
		"""Test parsing response that is a single user dict with uid."""
		response = {"name": "user1", "uid": "uid1"}
		result = self.workflow._parse_users_response(response)
		self.assertEqual(len(result), 1)
		self.assertEqual(result[0]["name"], "user1")

	def test_parse_users_response_empty_members_dict(self):
		"""Test parsing response with empty members dict."""
		response = {"members": {}}
		result = self.workflow._parse_users_response(response)
		self.assertEqual(len(result), 0)

	def test_parse_users_response_empty_members_list(self):
		"""Test parsing response with empty members list."""
		response = {"members": []}
		result = self.workflow._parse_users_response(response)
		self.assertEqual(len(result), 0)

	def test_parse_users_response_invalid_format(self):
		"""Test parsing response with invalid format."""
		response = {"data": "invalid"}
		result = self.workflow._parse_users_response(response)
		self.assertEqual(len(result), 0)

	# ===================================================================
	# DOMAIN PRIVILEGES TESTS
	# ===================================================================

	def test_create_user_multiple_domain_privileges(self):
		"""Test user creation with multiple domain privileges."""
		domain_privileges = [
			{"name": "domain1", "privilege": "browse"},
			{"name": "domain2", "privilege": "edit"},
			{"name": "domain3", "privilege": "super"}
		]
		self.workflow.create_user("testuser", "password123", domain_privileges)
		call_args = self.session_client.rest_client.post.call_args
		payload = call_args[0][1]
		self.assertEqual(len(payload["domainPrivileges"]), 3)
		self.assertEqual(payload["domainPrivileges"][0]["priv"], "browse")
		self.assertEqual(payload["domainPrivileges"][1]["priv"], "edit")
		self.assertEqual(payload["domainPrivileges"][2]["priv"], "super")

	def test_modify_user_multiple_domain_privileges(self):
		"""Test user modification with multiple domain privileges."""
		self._simulate_user_exists("modify_user", uid="uid_modify")
		domain_privileges = [
			{"name": "domain1", "privilege": "service"},
			{"name": "domain2", "privilege": "security_admin"}
		]
		self.workflow.modify_user_by_name("modify_user", domain_privileges=domain_privileges)
		call_args = self.session_client.rest_client.patch.call_args
		payload = call_args[0][1]
		self.assertEqual(len(payload["domainPrivileges"]), 2)
		self.assertEqual(payload["domainPrivileges"][0]["priv"], "service")
		self.assertEqual(payload["domainPrivileges"][1]["priv"], "security_admin")

	# ===================================================================
	# USER VALIDATOR TESTS
	# ===================================================================

	def test_validator_duplicate_domain_names(self):
		"""Test validator catches duplicate domain names."""
		domain_privileges = [
			{"name": "default", "privilege": "browse"},
			{"name": "default", "privilege": "edit"}
		]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", "password123", domain_privileges)
		self.assertIn("Duplicate domain names", str(context.exception))

	def test_validator_invalid_privilege_value(self):
		"""Test validator catches invalid privilege value."""
		domain_privileges = [{"name": "default", "privilege": "invalid_priv"}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", "password123", domain_privileges)
		self.assertIn("Invalid privilege", str(context.exception))

	def test_validator_password_list_invalid_characters(self):
		"""Test validator catches invalid characters in password list."""
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		password_list = ['p', 'a', 'ss', 'w']  # 'ss' is not a single character
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", password_list, domain_privileges)
		self.assertIn("single character", str(context.exception))

	def test_validator_password_list_empty(self):
		"""Test validator catches empty password list."""
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", [], domain_privileges)
		self.assertIn("Password cannot be empty", str(context.exception))

	def test_validator_domain_privilege_missing_name(self):
		"""Test validator catches domain privilege missing name field."""
		domain_privileges = [{"privilege": "browse"}]  # Missing 'name'
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", "password123", domain_privileges)
		self.assertIn("missing 'name' field", str(context.exception))

	def test_validator_domain_privilege_missing_privilege(self):
		"""Test validator catches domain privilege missing privilege field."""
		domain_privileges = [{"name": "default"}]  # Missing 'privilege'
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", "password123", domain_privileges)
		self.assertIn("missing 'privilege' field", str(context.exception))

	def test_validator_domain_privilege_not_dict(self):
		"""Test validator catches domain privilege not being a dict."""
		domain_privileges = ["browse"]  # Should be dict
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", "password123", domain_privileges)
		self.assertIn("must be a dictionary", str(context.exception))

	def test_validator_domain_name_empty(self):
		"""Test validator catches empty domain name."""
		domain_privileges = [{"name": "", "privilege": "browse"}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", "password123", domain_privileges)
		self.assertIn("cannot be empty", str(context.exception))

	def test_validator_domain_name_not_string(self):
		"""Test validator catches non-string domain name."""
		domain_privileges = [{"name": 123, "privilege": "browse"}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", "password123", domain_privileges)
		self.assertIn("must be a string", str(context.exception))

	def test_validator_privilege_not_string(self):
		"""Test validator catches non-string privilege value."""
		domain_privileges = [{"name": "default", "privilege": 123}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", "password123", domain_privileges)
		self.assertIn("must be a string", str(context.exception))

	def test_validator_same_passwords(self):
		"""Test validator catches same current and new password."""
		self._simulate_user_exists("modify_user", uid="uid_modify")
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.modify_user_by_name("modify_user", 
				current_password="samepass123", 
				new_password="samepass123")
		self.assertIn("must be different", str(context.exception))

	def test_validator_password_non_string_non_list(self):
		"""Test validator catches password that is neither string nor list."""
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		with self.assertRaises(exceptions.InvalidInput) as context:
			self.workflow.create_user("testuser", 12345, domain_privileges)
		self.assertIn("must be a string or list", str(context.exception))

	def test_validator_all_valid_privileges(self):
		"""Test all valid privilege values are accepted."""
		valid_privileges = ['super', 'service', 'security_admin', 'edit', 'create', 'browse', 'basic_edit']
		for priv in valid_privileges:
			domain_privileges = [{"name": "default", "privilege": priv}]
			# Should not raise exception
			try:
				self.workflow.create_user(f"user_{priv}", "password123", domain_privileges)
			except exceptions.InvalidInput:
				self.fail(f"Valid privilege '{priv}' was rejected")

	# ===================================================================
	# UID VALIDATION TESTS
	# ===================================================================

	def test_fetch_user_uid_missing_uid(self):
		"""Test handling of user found but without UID."""
		self.session_client.rest_client.get.return_value = {
			"members": {
				"uid1": {"name": "testuser"}  # Missing 'uid' field
			}
		}
		with self.assertRaises(exceptions.HPEStorageException) as context:
			self.workflow._fetch_user_uid_by_name("testuser")
		self.assertIn("has no UID", str(context.exception))

	def test_fetch_user_uid_null_uid(self):
		"""Test handling of user with null UID."""
		self.session_client.rest_client.get.return_value = {
			"members": {
				"uid1": {"name": "testuser", "uid": None}
			}
		}
		with self.assertRaises(exceptions.HPEStorageException) as context:
			self.workflow._fetch_user_uid_by_name("testuser")
		self.assertIn("has no UID", str(context.exception))

	def test_fetch_user_uid_empty_uid(self):
		"""Test handling of user with empty string UID."""
		self.session_client.rest_client.get.return_value = {
			"members": {
				"uid1": {"name": "testuser", "uid": ""}
			}
		}
		with self.assertRaises(exceptions.HPEStorageException) as context:
			self.workflow._fetch_user_uid_by_name("testuser")
		self.assertIn("has no UID", str(context.exception))

	# ===================================================================
	# ERROR HANDLING TESTS
	# ===================================================================

	def test_get_all_users_rest_exception(self):
		"""Test get_all_users handling REST exceptions."""
		self.session_client.rest_client.get.side_effect = Exception("Connection error")
		with self.assertRaises(exceptions.HPEStorageException):
			self.workflow.get_all_users()

	def test_get_user_by_name_rest_exception(self):
		"""Test get_user_by_name handling REST exceptions."""
		self.session_client.rest_client.get.side_effect = Exception("API error")
		try:
			self.workflow.get_user_by_name("testuser")
			self.fail("Expected exception was not raised")
		except Exception as e:
			# Exception is raised but may not be wrapped in HPEStorageException
			self.assertIn("API error", str(e))

	def test_create_user_rest_exception(self):
		"""Test create_user handling REST exceptions."""
		domain_privileges = [{"name": "default", "privilege": "browse"}]
		self.session_client.rest_client.post.side_effect = Exception("Create failed")
		with self.assertRaises(Exception):
			self.workflow.create_user("testuser", "password123", domain_privileges)

	def test_modify_user_rest_exception(self):
		"""Test modify_user handling REST exceptions."""
		self._simulate_user_exists("modify_user", uid="uid_modify")
		self.session_client.rest_client.patch.side_effect = Exception("Modify failed")
		with self.assertRaises(Exception):
			self.workflow.modify_user_by_name("modify_user", 
				current_password="old123", new_password="new456")

	def test_delete_user_rest_exception(self):
		"""Test delete_user handling REST exceptions."""
		self._simulate_user_exists("delete_user", uid="uid_delete")
		self.session_client.rest_client.delete.side_effect = Exception("Delete failed")
		with self.assertRaises(Exception):
			self.workflow.delete_user_by_name("delete_user")

	# ===================================================================
	# VALIDATOR HELPER FUNCTION TESTS
	# ===================================================================

	def test_validate_password_change_params_new_without_current(self):
		"""Test password change validation when new password provided without current."""
		with self.assertRaises(exceptions.InvalidInput) as context:
			validate_password_change_params(None, "newpass123")
		self.assertIn("currentPassword is required", str(context.exception))

	def test_validate_password_change_params_current_without_new(self):
		"""Test password change validation when current password provided without new."""
		with self.assertRaises(exceptions.InvalidInput) as context:
			validate_password_change_params("current123", None)
		self.assertIn("newPassword is required", str(context.exception))

	def test_validate_password_change_params_both_provided(self):
		"""Test password change validation when both passwords provided."""
		# Should not raise exception
		validate_password_change_params("currentpass123", "newpass456")

	def test_validate_password_change_params_same_passwords(self):
		"""Test password change validation when passwords are the same."""
		with self.assertRaises(exceptions.InvalidInput) as context:
			validate_password_change_params("samepass123", "samepass123")
		self.assertIn("must be different", str(context.exception))

	def test_validate_user_operation_params_invalid_operation(self):
		"""Test operation validator with invalid operation type."""
		with self.assertRaises(exceptions.InvalidInput) as context:
			validate_user_operation_params("invalid_op")
		self.assertIn("Invalid operation", str(context.exception))

	def test_validate_user_operation_params_create_missing_name(self):
		"""Test operation validator for create without name."""
		with self.assertRaises(exceptions.InvalidInput) as context:
			validate_user_operation_params("create", password="pass123", domain_privileges=[])
		self.assertIn("'name' is required", str(context.exception))

	def test_validate_user_operation_params_create_missing_password(self):
		"""Test operation validator for create without password."""
		with self.assertRaises(exceptions.InvalidInput) as context:
			validate_user_operation_params("create", name="user1", domain_privileges=[])
		self.assertIn("'password' is required", str(context.exception))

	def test_validate_user_operation_params_create_missing_domain_privileges(self):
		"""Test operation validator for create without domain_privileges."""
		with self.assertRaises(exceptions.InvalidInput) as context:
			validate_user_operation_params("create", name="user1", password="pass123")
		self.assertIn("'domain_privileges' is required", str(context.exception))

	def test_validate_user_operation_params_create_success(self):
		"""Test operation validator for create with all params."""
		# Should not raise exception
		validate_user_operation_params("create", name="user1", password="pass123", domain_privileges=[])

	def test_validate_user_operation_params_modify_no_identifier(self):
		"""Test operation validator for modify without uid or name."""
		with self.assertRaises(exceptions.InvalidInput) as context:
			validate_user_operation_params("modify", current_password="pass123")
		self.assertIn("'uid' or 'name' is required", str(context.exception))

	def test_validate_user_operation_params_modify_no_changes(self):
		"""Test operation validator for modify without modifiable params."""
		with self.assertRaises(exceptions.InvalidInput) as context:
			validate_user_operation_params("modify", name="user1")
		self.assertIn("At least one of the following parameters", str(context.exception))

	def test_validate_user_operation_params_modify_with_uid(self):
		"""Test operation validator for modify with uid."""
		# Should not raise exception
		validate_user_operation_params("modify", uid="uid123", current_password="pass123")

	def test_validate_user_operation_params_modify_with_name(self):
		"""Test operation validator for modify with name."""
		# Should not raise exception
		validate_user_operation_params("modify", name="user1", new_password="newpass", current_password="old")

	def test_validate_user_operation_params_delete_no_identifier(self):
		"""Test operation validator for delete without uid or name."""
		with self.assertRaises(exceptions.InvalidInput) as context:
			validate_user_operation_params("delete")
		self.assertIn("'uid' or 'name' is required", str(context.exception))

	def test_validate_user_operation_params_delete_with_uid(self):
		"""Test operation validator for delete with uid."""
		# Should not raise exception
		validate_user_operation_params("delete", uid="uid123")

	def test_validate_user_operation_params_delete_with_name(self):
		"""Test operation validator for delete with name."""
		# Should not raise exception
		validate_user_operation_params("delete", name="user1")

	def test_validate_user_operation_params_get_no_identifier(self):
		"""Test operation validator for get without uid or name."""
		with self.assertRaises(exceptions.InvalidInput) as context:
			validate_user_operation_params("get")
		self.assertIn("'uid' or 'name' is required", str(context.exception))

	def test_validate_user_operation_params_get_with_uid(self):
		"""Test operation validator for get with uid."""
		# Should not raise exception
		validate_user_operation_params("get", uid="uid123")

	def test_validate_user_operation_params_get_with_name(self):
		"""Test operation validator for get with name."""
		# Should not raise exception
		validate_user_operation_params("get", name="user1")

	def test_validate_user_operation_params_get_all(self):
		"""Test operation validator for get_all (no params required)."""
		# Should not raise exception
		validate_user_operation_params("get_all")


if __name__ == '__main__':
	unittest.main()


#    (c) Copyright 2026 Hewlett Packard Enterprise Development LP
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from hpe_storage_flowkit_py.v3.src.core.exceptions import HPEStorageException, QosDoesNotExist, QosAlreadyExists
from hpe_storage_flowkit_py.v3.src.core.session import SessionManager  
from hpe_storage_flowkit_py.v3.src.core.logger import Logger
from hpe_storage_flowkit_py.v3.src.validators.qos_validator import validate_qos_params
from hpe_storage_flowkit_py.v3.src.utils.utils import handle_async_response
from hpe_storage_flowkit_py.v3.src.workflows.task import TaskManager

logger = Logger()

class QosWorkflow:


    def __init__(self, session_mgr: SessionManager, task_manager: TaskManager):
        self.session_mgr = session_mgr
        self.task_manager = task_manager


    def _execute_create_qos(self, vvs_name, qos, **kwargs):
        logger.info(f"Starting QoS rule creation for '{vvs_name}'")
        
        # Check if QoS rule already exists
        existing_qos = self.get_qos(vvs_name)
        if existing_qos:
            logger.error(f"QoS rule '{vvs_name}' already exists")
            raise QosAlreadyExists(name=vvs_name)
        
        # First, validate all parameters in the qos dict
        # Build params dict for validation including ALL keys from qos dict
        params_to_validate = {
            'targetName': vvs_name,
        }
        
        # Add all parameters from qos dict to validation
        params_to_validate.update(qos)
        
        # Merge kwargs into params for validation
        params_to_validate.update(kwargs)
        
        # Validate with params dict - this will catch invalid parameters like maxBW
        validate_qos_params(params_to_validate)
        
        # Now extract the valid parameters
        iopsMaxLimit = qos.get('iopsMaxLimit')  
        bandwidthMaxLimitKiB = qos.get('bandwidthMaxLimitKiB')
        targetType = qos.get('targetType', 'QOS_TGT_VVSET')  # Default to VVSET if not provided
        
        # Validate that at least one limit is provided
        if not iopsMaxLimit and not bandwidthMaxLimitKiB:
            raise ValueError("At least one of iopsMaxLimit or bandwidthMaxLimitKiB must be provided for QoS rule")
        
        # Build payload
        qosRule = {}
        qosRule["targetName"] = vvs_name  
        qosRule["targetType"] = targetType
        if iopsMaxLimit:
            qosRule['iopsMaxLimit'] = int(iopsMaxLimit)
        if bandwidthMaxLimitKiB:
            qosRule['bandwidthMaxLimitKiB'] = int(bandwidthMaxLimitKiB)
        
        # Add any additional valid parameters from qos dict (like enable, allowAIQoS, etc.)
        for key in ['enable', 'allowAIQoS']:
            if key in qos:
                qosRule[key] = qos[key]
        
        # Add any additional kwargs to payload
        qosRule.update(kwargs)

        logger.info(f"Creating QOS rule with payload: {qosRule}")
        try:
            resp = self.session_mgr.rest_client.post("/qosconfigs", qosRule)
            logger.info(f"QoS creation initiated: {resp}")
            
            result = handle_async_response(self.task_manager, "QoS creation", vvs_name, resp)
            logger.info(f"QoS rule for '{vvs_name}' created successfully")
            return result
        except Exception as e:
            logger.exception(f"Error creating QOS rule {qosRule}: {str(e) or repr(e)}")
            return e

    def create_qos(self, vvs_name, qos, **kwargs):
        logger.info(f">>>>>>>Entered create_qos: vvs_name='{vvs_name}'")
        try:
            logger.info(f"Calling _execute_create_qos with qos={qos}, vvs_name={vvs_name}, kwargs={kwargs}")
            response = self._execute_create_qos(vvs_name, qos, **kwargs)
            if isinstance(response, Exception):
                raise response
            logger.info(f"create_qos response: {response}")
            return response
        except Exception as e:
            logger.exception(f"Failed to create QoS due to error: {e}")
            raise
        finally:
            logger.info(f"<<<<<<<Exited create_qos: vvs_name='{vvs_name}'")

    def _execute_modify_qos(self, name, **kwargs):
        logger.info(f"Starting QoS rule modification for '{name}'")
        # Build params dict for validation
        params_to_validate = {}
        # Merge kwargs into params for validation
        params_to_validate.update(kwargs)
        
        # Validate with params dict
        validate_qos_params(params_to_validate)
        
        # Get QoS UID with proper error handling
        resp_list = self.get_qos(name=name)
        if not resp_list:
            logger.error(f"QoS rule not found for modification: {name}")
            raise QosDoesNotExist(name=name)
        
        uid = resp_list[0].get("uid")
        if not uid:
            logger.error(f"QoS UID missing for: {name} in response {resp_list}")
            raise ValueError(f"QoS UID missing for: {name}")
        
        # Build payload from params and kwargs
        payload = {}
        
        # Add any additional kwargs to payload
        payload.update(kwargs)
        
        logger.info(f"Modifying QOS rule with payload: {payload}")
        try:
            response = self.session_mgr.rest_client.patch(f"/qosconfigs/{uid}", payload)
            logger.info(f"QoS modification initiated: {response}")
            
            result = handle_async_response(self.task_manager, "QoS modification", name, response)
            logger.info(f"QoS rule '{name}' modified successfully")
            return result
        except Exception as e:
            logger.exception(f"Error modifying QOS rule {payload}: {str(e) or repr(e)}")
            return e
        
    def modify_qos(self, name, **kwargs):
        logger.info(f">>>>>>>Entered modify_qos: name='{name}'")
        try:
            logger.info(f"Calling _execute_modify_qos with name={name}, kwargs={kwargs}")
            response = self._execute_modify_qos(name, **kwargs)
            if isinstance(response, Exception):
                raise response
            logger.info(f"modify_qos response: {response}")
            return response
        except Exception as e:
            logger.exception(f"Failed to modify QoS due to error: {e}")
            raise
        finally:
            logger.info(f"<<<<<<<Exited modify_qos: name='{name}'")
    
    def _execute_delete_qos(self, name):
        logger.info(f"Starting QoS rule deletion for '{name}'")
        # Get QoS UID with proper error handling
        resp_list = self.get_qos(name)
        if not resp_list:
            logger.error(f"QoS rule not found for deletion: {name}")
            raise QosDoesNotExist(name=name)
        
        uid = resp_list[0].get("uid")
        if not uid:
            logger.error(f"QoS UID missing for: {name} in response {resp_list}")
            raise ValueError(f"QoS UID missing for: {name}")
        
        logger.info(f"Deleting QOS with uid={uid}")
        try:
            response = self.session_mgr.rest_client.delete(f"/qosconfigs/{uid}")
            logger.info(f"QoS deletion initiated: {response}")
            
            result = handle_async_response(self.task_manager, "QoS deletion", name, response)
            logger.info(f"QoS rule '{name}' deleted successfully")
            return result
        except Exception as e:
            logger.exception(f"Error deleting qos rule with uid {uid}: {str(e) or repr(e)}")
            return e
    
    def delete_qos(self, name):
        logger.info(f">>>>>>>Entered delete_qos: name='{name}'")
        try:
            logger.info(f"Calling _execute_delete_qos with name={name}")
            response = self._execute_delete_qos(name)
            if isinstance(response, Exception):
                raise response
            logger.info(f"delete_qos response: {response}")
            return response
        except Exception as e:
            logger.exception(f"Failed to delete QoS due to error: {e}")
            raise
        finally:
            logger.info(f"<<<<<<<Exited delete_qos: name='{name}'")

    def get_qos(self, name):
        try:
            headers = {"experimentalfilter": "true"}
            endpoint = f"/qosconfigs?targetName={name}"
            logger.info(f"Getting QOS with endpoint={endpoint}, headers={headers}")
            response = self.session_mgr.rest_client.get(endpoint, headers=headers)
            logger.info(f"get_qos response: {response}")
            return response
        except HPEStorageException as e:
            logger.exception(f"HPEStorageException in get_qos: {str(e) or repr(e)}")
            raise

    def list_qos(self):
        try:
            logger.info("Listing all QOS configs")
            resp = self.session_mgr.rest_client.get("/qosconfigs")
            logger.info(f"list_qos response: {resp}")
            return resp.get("members", [])
        except HPEStorageException as e:
            logger.exception(f"HPEStorageException in list_qos: {str(e) or repr(e)}")
            raise
